# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    PlacePage - Place index page and individual Place pages
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from collections import defaultdict
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import PlaceType, Place, PlaceName, Media
from gramps.gen.plug.report import Bibliography
from gramps.gen.mime import is_image_type
from gramps.plugins.lib.libhtml import Html
from gramps.gen.utils.place import conv_lat_lon, coord_formats
from gramps.gen.utils.location import get_main_location
from gramps.gen.display.place import displayer as _pd

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import (
    alphabet_navigation,
    GOOGLE_MAPS,
    FULLCLEAR,
    MARKER_PATH,
    OPENLAYER,
    OSM_MARKERS,
    STAMEN_MARKERS,
    MARKERS,
    html_escape,
    sort_places,
    AlphabeticIndex,
)

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


######################################################
#                                                    #
#                    Place Pages                     #
#                                                    #
######################################################
class PlacePages(BasePage):
    """
    This class is responsible for displaying information about the 'Person'
    database objects. It displays this information under the 'Events'
    tab. It is told by the 'add_instances' call which 'Person's to display,
    and remembers the list of persons. A single call to 'display_pages'
    displays both the Event List (Index) page and all the Event
    pages.

    The base class 'BasePage' is initialised once for each page that is
    displayed.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)
        self.place_dict = defaultdict(set)
        self.placemappages = None
        self.mapservice = None
        self.person = None
        self.familymappages = None
        self.googlemapkey = None
        self.stamenopts = None

        # Place needs to display coordinates?
        self.display_coordinates = report.options["coordinates"]

    def display_pages(self, the_lang, the_title):
        """
        Generate and output the pages under the Place tab, namely the place
        index and the individual place pages.

        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        LOG.debug("obj_dict[Place]")
        for item in self.report.obj_dict[Place].items():
            LOG.debug("    %s", str(item))
        message = _("Creating place pages")
        progress_title = self.report.pgrs_title(the_lang)
        if self.report.options["inc_uplaces"]:
            place_count = len(self.r_db.get_place_handles())
        else:
            place_count = len(self.report.obj_dict[Place])
        with self.r_user.progress(progress_title, message, place_count + 1) as step:
            if self.report.options["inc_uplaces"]:
                # add unused place
                place_list = self.r_db.get_place_handles()
                for place_ref in place_list:
                    if place_ref not in self.report.obj_dict[Place]:
                        place = self.r_db.get_place_from_handle(place_ref)
                        if place:
                            place_name = place.get_title()
                            p_fname = self.report.build_url_fname(
                                place_ref, "plc", False, init=True
                            )
                            p_fname += self.ext
                            plc_dict = (p_fname, place_name, place.gramps_id, None)
                            self.report.obj_dict[Place][place_ref] = plc_dict
                            p_name = _pd.display(self.r_db, place, fmt=0)
                            cplace_name = p_name.split()[-1]
                            if len(place_name.split()) > 1:
                                splace_name = place_name.split()[-2]
                            else:
                                splace_name = cplace_name
                            plc_dict = (
                                place_ref,
                                p_name,
                                splace_name,
                                cplace_name,
                                place.gramps_id,
                                None,
                            )
                            self.report.obj_dict[PlaceName][p_name] = plc_dict

        with self.r_user.progress(
            progress_title, message, len(self.report.obj_dict[Place]) + 1
        ) as step:
            index = 1
            for place_name in self.report.obj_dict[PlaceName].keys():
                step()
                p_handle = self.report.obj_dict[PlaceName][place_name]
                index += 1
                if isinstance(p_handle, tuple):
                    self.placepage(
                        self.report, the_lang, the_title, p_handle[0], place_name
                    )
            step()
        self.placelistpage(self.report, the_lang, the_title)

    def __output_place(
        self,
        ldatec,
        tbody,
        first_place,
        pname,
        sname,
        cname,
        place_handle,
        letter,
        bucket_link,
    ):
        place = None
        if place_handle:
            place = self.r_db.get_place_from_handle(place_handle)
        if place:
            if place.get_change_time() > ldatec:
                ldatec = place.get_change_time()
            nbrs = len(pname.split(","))
            plc_title = pname
            if nbrs == 3:
                plc_title = ", ".join(pname.split(",")[:1])
            elif nbrs == 4:
                plc_title = ", ".join(pname.split(",")[:2])
            elif nbrs == 5:
                plc_title = ", ".join(pname.split(",")[:3])
            elif nbrs == 6:
                plc_title = ", ".join(pname.split(",")[:4])
            main_location = get_main_location(self.r_db, place)
            if not plc_title or plc_title == " ":
                letter = "&nbsp;"
            trow = Html("tr")
            tbody += trow
            tcell = Html("td", class_="ColumnLetter", inline=True)
            trow += tcell
            if first_place:
                # or primary_difference(letter, prev_letter, self.rlocale):
                first_place = False
                # prev_letter = letter
                trow.attr = 'class = "BeginLetter"'
                ttle = self._("Places beginning " "with letter %s") % letter
                tcell += Html("a", letter, name=letter, title=ttle, id_=bucket_link)
            else:
                tcell += "&nbsp;"
            trow += Html(
                "td",
                self.place_link(place.get_handle(), plc_title, place.get_gramps_id()),
                class_="ColumnName",
            )
            trow.extend(
                Html("td", data or "&nbsp;", class_=colclass, inline=True)
                for (colclass, data) in [
                    # Use the two last field of a place
                    # We could have strange values if we
                    # have a placename like:
                    # city, province, country, mainland, planet...
                    # mainland and planet are custom types
                    ["ColumnState", sname],
                    ["ColumnCountry", cname],
                ]
            )
            if self.display_coordinates:
                tcell1 = Html("td", class_="ColumnLatitude", inline=True)
                tcell2 = Html("td", class_="ColumnLongitude", inline=True)
                trow += tcell1, tcell2
                if place.lat and place.long:
                    latitude, longitude = conv_lat_lon(
                        place.lat,
                        place.long,
                        coord_formats[self.report.options["coord_format"]],
                    )
                    tcell1 += latitude
                    tcell2 += longitude
                else:
                    tcell1 += "&nbsp;"
                    tcell2 += "&nbsp;"
        return (ldatec, first_place)

    def placelistpage(self, report, the_lang, the_title):
        """
        Create a place index

        @param: report        -- The instance of the main report class
                                 for this report
        @param: the_lang      -- The lang to process
        @param: the_title     -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)

        output_file, sio = self.report.create_file("places")
        result = self.write_header(self._("Places"))
        placelistpage, dummy_head, dummy_body, outerwrapper = result
        ldatec = 0

        # begin places division
        with Html("div", class_="content", id="Places") as placelist:
            outerwrapper += placelist

            # place list page message
            msg = self._(
                "This page contains an index of all the places in the "
                "database, sorted by their title. "
                "Clicking on a place&#8217;s "
                "title will take you to that place&#8217;s page."
            )
            placelist += Html("p", msg, id="description")

            # begin alphabet navigation
            # Assemble all the places
            index = AlphabeticIndex(self.rlocale)
            # self.report.obj_dict[PlaceName] is a dict with key place_name and
            # values (place_fname, place_name, place.gramps_id, event)
            for place_name, value in self.report.obj_dict[PlaceName].items():
                index.addRecord(place_name, value)

            # Extract the buckets from the index
            index_list = []
            index.resetBucketIterator()
            while index.nextBucket():
                if index.bucketRecordCount != 0:
                    index_list.append(index.bucketLabel)
            # Output the navigation
            alpha_nav = alphabet_navigation(index_list, self.rlocale, rtl=self.dir)
            if alpha_nav:
                placelist += alpha_nav

            # begin places table and table head
            with Html(
                "table", class_="infolist primobjlist placelist " + self.dir
            ) as table:
                placelist += table

                # begin table head
                thead = Html("thead")
                table += thead

                trow = Html("tr")
                thead += trow

                if self.display_coordinates:
                    trow.extend(
                        Html("th", label, class_=colclass, inline=True)
                        for (label, colclass) in [
                            [self._("Letter"), "ColumnLetter"],
                            [self._("Name", "Place Name"), "ColumnName"],
                            [self._("State/Province"), "ColumnState"],
                            [self._("Country"), "ColumnCountry"],
                            [self._("Latitude"), "ColumnLatitude"],
                            [self._("Longitude"), "ColumnLongitude"],
                        ]
                    )
                else:
                    trow.extend(
                        Html("th", label, class_=colclass, inline=True)
                        for (label, colclass) in [
                            [self._("Letter"), "ColumnLetter"],
                            [self._("Name", "Place Name"), "ColumnName"],
                            [self._("State/Province"), "ColumnState"],
                            [self._("Country"), "ColumnCountry"],
                        ]
                    )

                # begin table body
                tbody = Html("tbody")
                table += tbody

                # For each bucket, output the places in that bucket
                index.resetBucketIterator()
                output = []
                dup_index = 0
                while index.nextBucket():
                    if index.bucketRecordCount != 0:
                        bucket_letter = index.bucketLabel
                        bucket_link = bucket_letter
                        if bucket_letter in output:
                            bucket_link = "%s (%i)" % (bucket_letter, dup_index)
                            dup_index += 1
                        output.append(bucket_letter)
                        # Assemble all the places in this bucket into a dict for
                        # sorting
                        place_dict = dict()
                        while index.nextRecord():
                            place_name = index.recordName
                            value = index.recordData
                            place_dict[place_name] = value

                        handle_list = sort_places(self.r_db, place_dict, self.rlocale)
                        first_place = True
                        for pname, place_handle in handle_list:
                            if not pname:
                                continue
                            val = self.report.obj_dict[PlaceName][pname]
                            nbelem = len(val)
                            if val and nbelem > 3:
                                if isinstance(place_handle, tuple):
                                    place = self.r_db.get_place_from_handle(
                                        place_handle[0]
                                    )
                                else:
                                    place = self.r_db.get_place_from_handle(
                                        place_handle
                                    )
                                main_location = get_main_location(self.r_db, place)
                                sname = main_location.get(PlaceType.STATE, "")
                                cname = main_location.get(PlaceType.COUNTRY, "")
                            elif nbelem == 3:
                                cname = val[3]
                                sname = val[2]
                            else:
                                val = [""]
                                cname = ""
                                sname = ""
                            (ldatec, first_place) = self.__output_place(
                                ldatec,
                                trow,
                                first_place,
                                pname,
                                sname,
                                cname,
                                val[0],
                                bucket_letter,
                                bucket_link,
                            )

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(placelistpage, output_file, sio, ldatec)

    def placepage(self, report, the_lang, the_title, place_handle, place_name):
        """
        Create a place page

        @param: report       -- The instance of the main report class
                                for this report
        @param: the_lang     -- The lang to process
        @param: the_title    -- The title page related to the language
        @param: place_handle -- The handle for the place to add
        @param: place_name   -- The alternate place name
        """
        place = report.database.get_place_from_handle(place_handle)
        if not place:
            return
        BasePage.__init__(self, report, the_lang, the_title, place.get_gramps_id())
        self.bibli = Bibliography()
        ldatec = place.get_change_time()
        apname = _pd.display(self.r_db, place, fmt=0)

        # if place_name:  # != apname: # store only the primary named page
        output_file, sio = self.report.create_file(place_handle, "plc")
        self.uplink = True
        self.page_title = place_name
        (placepage, head, dummy_body, outerwrapper) = self.write_header(_("Places"))

        self.placemappages = self.report.options["placemappages"]
        self.mapservice = self.report.options["mapservice"]
        self.googlemapkey = self.report.options["googlemapkey"]
        self.stamenopts = self.report.options["stamenopts"]

        # begin PlaceDetail Division
        with Html("div", class_="content", id="PlaceDetail") as placedetail:
            outerwrapper += placedetail

            media_list = place.get_media_list()
            if media_list and self.create_media:
                if self.the_lang and not self.usecms:
                    fname = "/".join(["..", "css", "lightbox.css"])
                    jsname = "/".join(["..", "css", "lightbox.js"])
                else:
                    fname = "/".join(["css", "lightbox.css"])
                    jsname = "/".join(["css", "lightbox.js"])
                url = self.report.build_url_fname(fname, None, self.uplink)
                head += Html(
                    "link", href=url, type="text/css", media="screen", rel="stylesheet"
                )
                url = self.report.build_url_fname(jsname, None, self.uplink)
                head += Html("script", src=url, type="text/javascript", inline=True)
                thumbnail = self.disp_first_img_as_thumbnail(media_list, place)
                if thumbnail is not None:
                    if media_list[0].ref in self.report.obj_dict[Media]:
                        placedetail += thumbnail

            # add section title
            placedetail += Html("h3", html_escape(place_name), inline=True)

            # begin summaryarea division and places table
            with Html("div", id="summaryarea") as summaryarea:
                placedetail += summaryarea

                with Html("table", class_="infolist place") as table:
                    summaryarea += table

                    # list the place fields
                    self.dump_place(place, table)

            # place gallery
            if self.create_media and not self.report.options["inc_uplaces"]:
                # Don't diplay media for unused places. It generates
                # "page not found" if they are not collected in pass 1.
                placegallery = self.disp_add_img_as_gallery(media_list, place)
                if placegallery is not None:
                    placedetail += placegallery

            # place notes
            notelist = self.display_note_list(place.get_note_list(), Place)
            if notelist is not None:
                placedetail += notelist

            # place urls
            urllinks = self.display_url_list(place.get_url_list())
            if urllinks is not None:
                placedetail += urllinks

            # add place map here
            # Link to Gramps marker
            fname = "/".join(["images", "marker.png"])
            marker_path = self.report.build_url_image(
                "marker.png", "images", self.uplink
            )

            if self.placemappages:
                if place and (place.lat and place.long):
                    placetitle = place_name

                    # add narrative-maps CSS...
                    if the_lang and not self.usecms:
                        fname = "/".join(["..", "css", "narrative-maps.css"])
                    else:
                        fname = "/".join(["css", "narrative-maps.css"])
                    url = self.report.build_url_fname(fname, None, self.uplink)
                    head += Html(
                        "link",
                        href=url,
                        type="text/css",
                        media="screen",
                        rel="stylesheet",
                    )

                    # add MapService specific javascript code
                    src_js = GOOGLE_MAPS + "api/js"
                    if self.mapservice == "Google":
                        if self.googlemapkey:
                            src_js += "?key=" + self.googlemapkey
                        head += Html(
                            "script", type="text/javascript", src=src_js, inline=True
                        )
                    else:  # OpenStreetMap, Stamen...
                        src_js = self.secure_mode
                        src_js += (
                            "ajax.googleapis.com/ajax/libs/jquery/1.9.1/"
                            "jquery.min.js"
                        )
                        head += Html(
                            "script", type="text/javascript", src=src_js, inline=True
                        )
                        olv = self.report.options["ol_version"]
                        build = "legacy"
                        if olv < "v7.0.0":
                            build = "build"
                        if olv == "latest":
                            build = "legacy"
                        src_js = (
                            "https://openlayers.org/en/" "%(ver)s/%(bld)s/ol.js"
                        ) % {"ver": olv, "bld": build}
                        head += Html(
                            "script", type="text/javascript", src=src_js, inline=True
                        )
                        css = "legacy"
                        if olv < "v7.0.0":
                            css = "css"
                        if olv == "latest":
                            css = "legacy"
                        url = (
                            "https://openlayers.org/en/" "%(ver)s/%(css)s/ol.css"
                        ) % {"ver": olv, "css": css}
                        head += Html(
                            "link", href=url, type="text/css", rel="stylesheet"
                        )

                    # section title
                    placedetail += Html("h4", self._("Place Map"), inline=True)

                    # begin map_canvas division
                    with Html("div", id="map_canvas", inline=True) as canvas:
                        placedetail += canvas

            # add div for popups.
            if self.mapservice == "Google":
                with Html("div", id="popup", inline=True) as popup:
                    placedetail += popup
            else:
                with Html("div", id="popup", class_="ol-popup", inline=True) as popup:
                    placedetail += popup
                    popup += Html(
                        "a", href="#", id="popup-closer", class_="ol-popup-closer"
                    )
                    popup += Html("div", id="popup-title", class_="ol-popup-title")
                    popup += Html("div", id="popup-content")
                with Html(
                    "div", id="tooltip", class_="ol-popup", inline=True
                ) as tooltip:
                    placedetail += tooltip
                    tooltip += Html("div", id="tooltip-content")

            # source references
            if not self.report.options["inc_uplaces"]:
                # We can't display source reference when we display
                # unused places. These info are not in the collected objects.
                # This is to avoid "page not found" errors.
                srcrefs = self.display_ind_sources(place)
                if srcrefs is not None:
                    placedetail += srcrefs

            # References list
            ref_list = self.display_bkref_list(Place, place_handle)
            if ref_list is not None:
                placedetail += ref_list

            # Begin inline javascript code because jsc is a
            # docstring, it does NOT have to be properly indented
            latitude, longitude = conv_lat_lon(
                place.get_latitude(), place.get_longitude(), "D.D8"
            )
            if not (latitude and longitude):
                # We have incorrect longitude and/or latitude for this place.
                latitude = longitude = 0.0
            if self.placemappages:
                if place and (place.lat and place.long):
                    tracelife = " "
                    if self.create_media and media_list:
                        for fmedia in media_list:
                            photo_hdle = fmedia.get_reference_handle()
                            photo = self.r_db.get_media_from_handle(photo_hdle)
                            mime_type = photo.get_mime_type()
                            descr = photo.get_description()

                            if mime_type and is_image_type(mime_type):
                                uplnk = self.uplink
                                (pth, dummy_) = self.report.prepare_copy_media(photo)
                                srbuf = self.report.build_url_fname
                                newpath = srbuf(pth, image=True, uplink=uplnk)
                                imglnk = self.media_link(
                                    photo_hdle,
                                    newpath,
                                    descr,
                                    uplink=uplnk,
                                    usedescr=False,
                                )
                                if photo_hdle in self.report.obj_dict[Media]:
                                    tracelife += str(imglnk)
                                break  # We show only the first image
                    with Html("script", type="text/javascript", indent=False) as jsc:
                        if self.mapservice == "Google":
                            # scripts += jsc
                            # Google adds Latitude/ Longitude to its maps...
                            plce = placetitle.replace("'", "\\'")
                            jsc += MARKER_PATH % marker_path
                            jsc += MARKERS % (
                                [[plce, latitude, longitude, 1, tracelife]],
                                latitude,
                                longitude,
                                10,
                            )
                        elif self.mapservice == "OpenStreetMap":
                            jsc += MARKER_PATH % marker_path
                            jsc += OSM_MARKERS % (
                                [
                                    [
                                        float(longitude),
                                        float(latitude),
                                        placetitle,
                                        tracelife,
                                    ]
                                ],
                                longitude,
                                latitude,
                                10,
                            )
                            jsc += OPENLAYER
                        else:  # STAMEN
                            jsc += MARKER_PATH % marker_path
                            jsc += STAMEN_MARKERS % (
                                [
                                    [
                                        float(longitude),
                                        float(latitude),
                                        placetitle,
                                        tracelife,
                                    ]
                                ],
                                self.stamenopts,
                                longitude,
                                latitude,
                                10,
                            )
                            jsc += OPENLAYER
                    placedetail += jsc

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        # if place_name:  # != apname: # store only the primary named page
        self.xhtml_writer(placepage, output_file, sio, ldatec)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2012 Brian G. Matherly
# Copyright (C) 2009      Gary Burton
# Contribution  2009 by   Reinhard Mueller <reinhard.mueller@bytewise.at>
# Copyright (C) 2010      Jakim Friant
# Copyright (C) 2013-2014 Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Reports/Text Reports/Kinship Report"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.errors import ReportError
from gramps.gen.relationship import get_relationship_calculator
from gramps.gen.plug.docgen import (
    IndexMark,
    FontStyle,
    ParagraphStyle,
    FONT_SANS_SERIF,
    INDEX_TYPE_TOC,
    PARA_ALIGN_CENTER,
)
from gramps.gen.plug.menu import NumberOption, BooleanOption, PersonOption
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gen.proxy import CacheProxyDb
from gramps.gen.display.name import displayer as _nd


# ------------------------------------------------------------------------
#
# KinshipReport
#
# ------------------------------------------------------------------------
class KinshipReport(Report):
    """Kinship Report"""

    def __init__(self, database, options, user):
        """
        Create the KinshipReport object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        This report needs the following parameters (class variables)
        that come in the options class.

        maxdescend    - Maximum generations of descendants to include.
        maxascend     - Maximum generations of ancestors to include.
        incspouses    - Whether to include spouses.
        inccousins    - Whether to include cousins.
        incaunts      - Whether to include aunts/uncles/nephews/nieces.
        pid           - The Gramps ID of the center person for the report.
        name_format   - Preferred format to display names
        incl_private  - Whether to include private data
        living_people - How to handle living people
        years_past_death - Consider as living this many years after death
        """
        Report.__init__(self, database, options, user)
        menu = options.menu

        self.set_locale(menu.get_option_by_name("trans").get_value())

        stdoptions.run_date_format_option(self, menu)

        stdoptions.run_private_data_option(self, menu)
        stdoptions.run_living_people_option(self, menu, self._locale)
        self.database = CacheProxyDb(self.database)
        self.__db = self.database

        self.max_descend = menu.get_option_by_name("maxdescend").get_value()
        self.max_ascend = menu.get_option_by_name("maxascend").get_value()
        self.inc_spouses = menu.get_option_by_name("incspouses").get_value()
        self.inc_cousins = menu.get_option_by_name("inccousins").get_value()
        self.inc_aunts = menu.get_option_by_name("incaunts").get_value()
        pid = menu.get_option_by_name("pid").get_value()
        self.person = self.database.get_person_from_gramps_id(pid)
        if self.person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        stdoptions.run_name_format_option(self, menu)

        self.rel_calc = get_relationship_calculator(reinit=True, clocale=self._locale)

        self.kinship_map = {}
        self.spouse_map = {}

    def write_report(self):
        """
        The routine the actually creates the report. At this point, the document
        is opened and ready for writing.
        """
        pname = self._name_display.display(self.person)

        self.doc.start_paragraph("KIN-Title")
        # feature request 2356: avoid genitive form
        title = self._("Kinship Report for %s") % pname
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        self.doc.end_paragraph()

        if self.inc_spouses:
            spouse_handles = self.get_spouse_handles(self.person.get_handle())
            if spouse_handles:
                self.write_people(self._("Spouses"), spouse_handles)

        # Collect all descendants of the person
        self.traverse_down(self.person.get_handle(), 0, 1)

        # Collect all ancestors/aunts/uncles/nephews/cousins of the person
        self.traverse_up(self.person.get_handle(), 1, 0)

        # Write Kin
        for Ga, Gbs in self.kinship_map.items():
            for Gb in Gbs:
                # To understand these calculations, see:
                # http://en.wikipedia.org/wiki/Cousin#Mathematical_definitions
                _x_ = min(Ga, Gb)
                _y_ = abs(Ga - Gb)
                # Skip unrequested people
                if _x_ == 1 and _y_ > 0 and not self.inc_aunts:
                    continue
                elif _x_ > 1 and not self.inc_cousins:
                    continue

                get_rel_str = self.rel_calc.get_plural_relationship_string

                title = get_rel_str(Ga, Gb, in_law_b=False)
                self.write_people(self._(title), self.kinship_map[Ga][Gb])

                if (
                    self.inc_spouses
                    and Ga in self.spouse_map
                    and Gb in self.spouse_map[Ga]
                ):
                    title = get_rel_str(Ga, Gb, in_law_b=True)
                    self.write_people(self._(title), self.spouse_map[Ga][Gb])

    def traverse_down(self, person_handle, Ga, Gb, skip_handle=None):
        """
        Populate a map of arrays containing person handles for the descendants
        of the passed person. This function calls itself recursively until it
        reaches max_descend.

        Parameters:
        person_handle: the handle of the person to go to next
        Ga: The number of generations from the main person to the common
           ancestor. This should be incremented when going up generations, and
           left alone when going down generations.
        Gb: The number of generations from this person (person_handle) to the
           common ancestor. This should be incremented when going down
           generations and set back to zero when going up generations.
        skip_handle: an optional handle to skip when going down. This is useful
           to skip the descendant that brought you this generation in the first
           place.
        """
        for child_handle in self.get_children_handles(person_handle):
            if child_handle != skip_handle:
                self.add_kin(child_handle, Ga, Gb)

                if self.inc_spouses:
                    for spouse_handle in self.get_spouse_handles(child_handle):
                        self.add_spouse(spouse_handle, Ga, Gb)

                if Gb < self.max_descend:
                    self.traverse_down(child_handle, Ga, Gb + 1)

    def traverse_up(self, person_handle, Ga, Gb):
        """
        Populate a map of arrays containing person handles for the ancestors
        of the passed person. This function calls itself recursively until it
        reaches max_ascend.

        Parameters:
        person_handle: the handle of the person to go to next
        Ga: The number of generations from the main person to the common
           ancestor. This should be incremented when going up generations, and
           left alone when going down generations.
        Gb: The number of generations from this person (person_handle) to the
           common ancestor. This should be incremented when going down
           generations and set back to zero when going up generations.
        """
        parent_handles = self.get_parent_handles(person_handle)
        for parent_handle in parent_handles:
            self.add_kin(parent_handle, Ga, Gb)
            self.traverse_down(parent_handle, Ga, Gb + 1, person_handle)
            if Ga < self.max_ascend:
                self.traverse_up(parent_handle, Ga + 1, 0)

    def add_kin(self, person_handle, Ga, Gb):
        """
        Add a person handle to the kin map.
        """
        if Ga not in self.kinship_map:
            self.kinship_map[Ga] = {}
        if Gb not in self.kinship_map[Ga]:
            self.kinship_map[Ga][Gb] = []
        if person_handle not in self.kinship_map[Ga][Gb]:
            self.kinship_map[Ga][Gb].append(person_handle)

    def add_spouse(self, spouse_handle, Ga, Gb):
        """
        Add a person handle to the spouse map.
        """
        if Ga not in self.spouse_map:
            self.spouse_map[Ga] = {}
        if Gb not in self.spouse_map[Ga]:
            self.spouse_map[Ga][Gb] = []
        if spouse_handle not in self.spouse_map[Ga][Gb]:
            self.spouse_map[Ga][Gb].append(spouse_handle)

    def get_parent_handles(self, person_handle):
        """
        Return an array of handles for all the parents of the
        given person handle.
        """
        parent_handles = []
        person = self.__db.get_person_from_handle(person_handle)
        family_handle = person.get_main_parents_family_handle()
        if family_handle:
            family = self.__db.get_family_from_handle(family_handle)
            father_handle = family.get_father_handle()
            if father_handle:
                parent_handles.append(father_handle)
            mother_handle = family.get_mother_handle()
            if mother_handle:
                parent_handles.append(mother_handle)
        return parent_handles

    def get_spouse_handles(self, person_handle):
        """
        Return an array of handles for all the spouses of the
        given person handle.
        """
        spouses = []
        person = self.__db.get_person_from_handle(person_handle)
        for family_handle in person.get_family_handle_list():
            family = self.__db.get_family_from_handle(family_handle)
            father_handle = family.get_father_handle()
            mother_handle = family.get_mother_handle()
            spouse_handle = None
            if mother_handle and father_handle == person_handle:
                spouse_handle = mother_handle
            elif father_handle and mother_handle == person_handle:
                spouse_handle = father_handle

            if spouse_handle and spouse_handle not in spouses:
                spouses.append(spouse_handle)
        return spouses

    def get_children_handles(self, person_handle):
        """
        Return an array of handles for all the children of the
        given person handle.
        """
        children = []
        person = self.__db.get_person_from_handle(person_handle)
        for family_handle in person.get_family_handle_list():
            family = self.__db.get_family_from_handle(family_handle)
            for child_ref in family.get_child_ref_list():
                children.append(child_ref.get_reference_handle())
        return children

    def write_people(self, title, people_handles):
        """
        Write information about a group of people - including the title.
        """
        cap_title = title[0].upper() + title[1:]
        subtitle = "%s (%d)" % (cap_title, len(people_handles))
        self.doc.start_paragraph("KIN-Subtitle")
        mark = IndexMark(cap_title, INDEX_TYPE_TOC, 2)
        self.doc.write_text(subtitle, mark)
        self.doc.end_paragraph()
        list(map(self.write_person, people_handles))

    def write_person(self, person_handle):
        """
        Write information about the given person.
        """
        person = self.database.get_person_from_handle(person_handle)

        name = self._name_display.display(person)
        mark = utils.get_person_mark(self.database, person)
        birth_date = ""
        birth = get_birth_or_fallback(self.database, person)
        if birth:
            birth_date = self._get_date(birth.get_date_object())

        death_date = ""
        death = get_death_or_fallback(self.database, person)
        if death:
            death_date = self._get_date(death.get_date_object())
        dates = ""
        if birth_date or death_date:
            dates = " (%(birth_date)s - %(death_date)s)" % {
                "birth_date": birth_date,
                "death_date": death_date,
            }

        self.doc.start_paragraph("KIN-Normal")
        self.doc.write_text(name, mark)
        self.doc.write_text(dates)
        self.doc.end_paragraph()


# ------------------------------------------------------------------------
#
# KinshipOptions
#
# ------------------------------------------------------------------------
class KinshipOptions(MenuReportOptions):
    """
    Defines options and provides handling interface.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        return _nd.display(person)

    def add_menu_options(self, menu):
        """
        Add options to the menu for the kinship report.
        """
        category_name = _("Report Options")

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the report"))
        menu.add_option(category_name, "pid", self.__pid)

        maxdescend = NumberOption(_("Max Descendant Generations"), 2, 1, 50)
        maxdescend.set_help(_("The maximum number of descendant generations"))
        menu.add_option(category_name, "maxdescend", maxdescend)

        maxascend = NumberOption(_("Max Ancestor Generations"), 2, 1, 50)
        maxascend.set_help(_("The maximum number of ancestor generations"))
        menu.add_option(category_name, "maxascend", maxascend)

        incspouses = BooleanOption(_("Include spouses"), True)
        incspouses.set_help(_("Whether to include spouses"))
        menu.add_option(category_name, "incspouses", incspouses)

        inccousins = BooleanOption(_("Include cousins"), True)
        inccousins.set_help(_("Whether to include cousins"))
        menu.add_option(category_name, "inccousins", inccousins)

        incaunts = BooleanOption(_("Include aunts/uncles/nephews/nieces"), True)
        incaunts.set_help(_("Whether to include aunts/uncles/nephews/nieces"))
        menu.add_option(category_name, "incaunts", incaunts)

        category_name = _("Report Options (2)")

        stdoptions.add_name_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_living_people_option(menu, category_name)

        locale_opt = stdoptions.add_localization_option(menu, category_name)

        stdoptions.add_date_format_option(menu, category_name, locale_opt)

    def make_default_style(self, default_style):
        """Make the default output style for the Kinship Report."""
        font = FontStyle()
        font.set_size(16)
        font.set_type_face(FONT_SANS_SERIF)
        font.set_bold(1)
        para = ParagraphStyle()
        para.set_header_level(1)
        para.set_bottom_border(1)
        para.set_bottom_margin(utils.pt2cm(8))
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("KIN-Title", para)

        font = FontStyle()
        font.set_size(12)
        font.set_bold(True)
        para = ParagraphStyle()
        para.set_header_level(3)
        para.set_font(font)
        para.set_top_margin(utils.pt2cm(6))
        para.set_description(_("The style used for second level headings."))
        default_style.add_paragraph_style("KIN-Subtitle", para)

        font = FontStyle()
        font.set_size(10)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_left_margin(0.5)
        para.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("KIN-Normal", para)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009-2011  Gary Burton
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
EditPerson Dialog. Provide the interface to allow the Gramps program
to edit information about a particular Person.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from copy import copy
import pickle

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository.GLib import markup_escape_text

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.utils.file import media_path_full
from gramps.gen.utils.thumbnails import get_thumbnail_image
from ..utils import is_right_click, open_file_with_default_application
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gen.lib import NoteType, Person, Surname
from gramps.gen.db import DbTxn
from .. import widgets
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.errors import WindowActiveError
from ..glade import Glade
from ..ddtargets import DdTargets
from ..widgets.menuitem import add_menuitem

from .editprimary import EditPrimary
from .editmediaref import EditMediaRef
from .editname import EditName
from gramps.gen.config import config
from ..dialog import ErrorDialog, ICON
from gramps.gen.errors import ValidationError

from .displaytabs import (
    PersonEventEmbedList,
    NameEmbedList,
    CitationEmbedList,
    AttrEmbedList,
    AddrEmbedList,
    NoteTab,
    GalleryTab,
    WebEmbedList,
    PersonRefEmbedList,
    LdsEmbedList,
    PersonBackRefList,
    SurnameTab,
)
from gramps.gen.plug import CATEGORY_QR_PERSON
from gramps.gen.const import URL_MANUAL_SECT1
from gramps.gen.utils.id import create_id

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT1

_select_gender = ((True, False, False), (False, True, False), (False, False, True))


class SingSurn:
    """
    Managing the single surname components
    """

    def __init__(self, gladeobj):
        self.top = gladeobj

    def hide_all(self):
        # self.top.get_object('prefixlabel').hide()
        self.top.get_object("prefix").hide()
        self.top.get_object("surnamelabel").hide()
        self.top.get_object("surname").hide()
        self.top.get_object("originlabel").hide()
        self.top.get_object("originlabel").hide()
        self.top.get_object("cmborigin").hide()
        self.top.get_object("multsurnamebtn").hide()

    def show_all(self):
        # self.top.get_object('prefixlabel').show()
        self.top.get_object("prefix").show()
        self.top.get_object("surnamelabel").show()
        self.top.get_object("surname").show()
        self.top.get_object("originlabel").show()
        self.top.get_object("originlabel").show()
        self.top.get_object("cmborigin").show()
        self.top.get_object("multsurnamebtn").show()


class EditPerson(EditPrimary):
    """
    The EditPerson dialog is derived from the EditPrimary class.

    It allows for the editing of the primary object type of Person.

    """

    QR_CATEGORY = CATEGORY_QR_PERSON

    def __init__(self, dbstate, uistate, track, person, callback=None):
        """
        Create an EditPerson window.

        Associate a person with the window.

        """
        EditPrimary.__init__(
            self,
            dbstate,
            uistate,
            track,
            person,
            dbstate.db.get_person_from_handle,
            dbstate.db.get_person_from_gramps_id,
            callback,
        )

    def empty_object(self):
        """
        Return an empty Person object for comparison for changes.

        This is used by the base class (EditPrimary).

        """
        person = Person()
        # the editor requires a surname
        person.primary_name.add_surname(Surname())
        person.primary_name.set_primary_surname(0)
        return person

    def get_menu_title(self):
        if self.obj and self.obj.get_handle():
            name = name_displayer.display(self.obj)
            title = _("Person: %(name)s") % {"name": name}
        else:
            name = name_displayer.display(self.obj)
            if name:
                title = _("New Person: %(name)s") % {"name": name}
            else:
                title = _("New Person")
        return title

    def get_preview_name(self):
        prevname = name_displayer.display(self.obj)
        return prevname

    def _local_init(self):
        """
        Performs basic initialization, including setting up widgets and the
        glade interface.

        Local initialization function.
        This is called by the base class of EditPrimary, and overridden here.

        """
        self.pname = self.obj.get_primary_name()
        self.should_guess_gender = (
            not self.obj.get_gramps_id() and self.obj.get_gender() == Person.UNKNOWN
        )

        self.added = self.obj.handle is None
        if self.added:
            self.obj.handle = create_id()

        self.top = Glade()

        self.set_window(self.top.toplevel, None, self.get_menu_title())
        self.setup_configs("interface.person", 750, 550)

        self.obj_photo = self.top.get_object("personPix")
        self.frame_photo = self.top.get_object("frame5")
        self.eventbox = self.top.get_object("eventbox1")
        self.singsurnfr = SingSurn(self.top)
        self.multsurnfr = self.top.get_object("hboxmultsurnames")
        self.singlesurn_active = True

        self.set_contexteventbox(self.top.get_object("eventboxtop"))

    def _post_init(self):
        """
        Handle any initialization that needs to be done after the interface is
        brought up.

        Post initalization function.
        This is called by _EditPrimary's init routine, and overridden in the
        derived class (this class).

        """
        self.load_person_image()
        self.given.grab_focus()
        self._changed_name(None)
        self.top.get_object("hboxmultsurnames").pack_start(self.surntab, True, True, 0)

        if len(self.obj.get_primary_name().get_surname_list()) > 1:
            self.multsurnfr.set_size_request(
                -1, int(config.get("interface.surname-box-height"))
            )
            self.singsurnfr.hide_all()
            self.multsurnfr.show_all()
            self.singlesurn_active = False
        else:
            self.multsurnfr.hide()
            self.singsurnfr.show_all()
            self.singlesurn_active = True
        # if self.pname.get_surname() and not self.pname.get_first_name():
        #    self.given.grab_focus()
        # else:
        #    self.surname_field.grab_focus()

    def _connect_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self.define_cancel_button(self.top.get_object("button15"))
        self.define_ok_button(self.top.get_object("ok"), self.save)
        self.define_help_button(
            self.top.get_object("button134"),
            WIKI_HELP_PAGE,
            _("Editing_information_about_people", "manual"),
        )

        self.given.connect("focus-out-event", self._given_focus_out_event)
        self.top.get_object("editnamebtn").connect("clicked", self._edit_name_clicked)
        self.top.get_object("multsurnamebtn").connect(
            "clicked", self._mult_surn_clicked
        )

        self.eventbox.connect("button-press-event", self._image_button_press)
        # allow to initiate a drag-and-drop with this person if it has a handle
        if self.added:
            return  # Avoid HandleError if dragging an objet not in db yet
        self.contexteventbox.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            [DdTargets.PERSON_LINK.target()],
            Gdk.DragAction.COPY,
        )
        self.contexteventbox.drag_source_set_icon_name("gramps-person")
        self.contexteventbox.connect("drag_data_get", self._top_drag_data_get)

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("person-rebuild", self._do_close)
        self._add_db_signal("person-delete", self.check_for_close)
        self._add_db_signal("family-rebuild", self.family_change)
        self._add_db_signal("family-delete", self.family_change)
        self._add_db_signal("family-update", self.family_change)
        self._add_db_signal("family-add", self.family_change)
        self._add_db_signal("event-rebuild", self.event_updated)

    def family_change(self, handle_list=[]):
        """
        Callback for family change signals.

        This should rebuild the
           backreferences to family in person when:
            1)a family the person is parent of changes. Person could have
              been removed
            2)a family the person is child in changes. Child could have been
              removed
            3)a family is changed. The person could be added as child or
              parent

        """
        # As this would be an extensive check, we choose the easy path and
        #   rebuild family backreferences on all family changes
        self._update_families()

    def _update_families(self):
        phandle = self.obj.get_handle()
        if self.dbstate.db.has_person_handle(phandle):
            # new person has no handle yet and cannot be in a family.
            person = self.dbstate.db.get_person_from_handle(phandle)
            self.obj.set_family_handle_list(person.get_family_handle_list())
            self.obj.set_parent_family_handle_list(
                person.get_parent_family_handle_list()
            )
            # a family groupname in event_list might need to be changed
            # we just rebuild the view always
            self.event_list.rebuild_callback()

    def event_updated(self, obj):
        # place in event might have changed, or person event shown in the list
        # we just rebuild the view always
        self.event_list.rebuild_callback()

    def _validate_call(self, widget, text):
        """a callname must be a part of the given name, see if this is the
        case"""
        validcall = self.given.obj.get_text().split()
        dummy = copy(validcall)
        for item in dummy:
            validcall += item.split("-")
        if text in validcall:
            return
        return ValidationError(
            _("Call name must be the given name that " "is normally used.")
        )

    def _setup_fields(self):
        """
        Connect the GrampsWidget objects to field in the interface.

        This allows the widgets to keep the data in the attached Person object
        up to date at all times, eliminating a lot of need in 'save' routine.

        """

        self.private = widgets.PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

        self.gender = widgets.MonitoredMenu(
            self.top.get_object("gender"),
            self.obj.set_gender,
            self.obj.get_gender,
            (
                (_("female"), Person.FEMALE),
                (_("male"), Person.MALE),
                (_("unknown", "gender"), Person.UNKNOWN),
                (_("other", "gender"), Person.OTHER),
            ),
            self.db.readonly,
        )

        self.ntype_field = widgets.MonitoredDataType(
            self.top.get_object("ntype"),
            self.pname.set_type,
            self.pname.get_type,
            self.db.readonly,
            self.db.get_name_types(),
        )

        # part of Given Name section
        self.given = widgets.MonitoredEntry(
            self.top.get_object("given_name"),
            self.pname.set_first_name,
            self.pname.get_first_name,
            self.db.readonly,
        )

        self.call = widgets.MonitoredEntry(
            self.top.get_object("call"),
            self.pname.set_call_name,
            self.pname.get_call_name,
            self.db.readonly,
        )
        self.call.connect("validate", self._validate_call)
        # force validation now with initial entry
        self.call.obj.validate(force=True)

        self.title = widgets.MonitoredEntry(
            self.top.get_object("title"),
            self.pname.set_title,
            self.pname.get_title,
            self.db.readonly,
        )

        self.suffix = widgets.MonitoredEntryIndicator(
            self.top.get_object("suffix"),
            self.pname.set_suffix,
            self.pname.get_suffix,
            _("suffix"),
            self.db.readonly,
        )

        self.nick = widgets.MonitoredEntry(
            self.top.get_object("nickname"),
            self.pname.set_nick_name,
            self.pname.get_nick_name,
            self.db.readonly,
        )

        # part of Single Surname section
        self.surname_field = widgets.MonitoredEntry(
            self.top.get_object("surname"),
            self.pname.get_primary_surname().set_surname,
            self.pname.get_primary_surname().get_surname,
            self.db.readonly,
            autolist=self.db.get_surname_list() if not self.db.readonly else [],
        )

        self.prefix = widgets.MonitoredEntryIndicator(
            self.top.get_object("prefix"),
            self.pname.get_primary_surname().set_prefix,
            self.pname.get_primary_surname().get_prefix,
            _("prefix"),
            self.db.readonly,
        )

        self.ortype_field = widgets.MonitoredDataType(
            self.top.get_object("cmborigin"),
            self.pname.get_primary_surname().set_origintype,
            self.pname.get_primary_surname().get_origintype,
            self.db.readonly,
            self.db.get_origin_types(),
        )

        # other fields

        self.tags = widgets.MonitoredTagList(
            self.top.get_object("tag_label"),
            self.top.get_object("tag_button"),
            self.obj.set_tag_list,
            self.obj.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.gid = widgets.MonitoredEntry(
            self.top.get_object("gid"),
            self.obj.set_gramps_id,
            self.obj.get_gramps_id,
            self.db.readonly,
        )

        # make sure title updates automatically
        for obj in [
            self.top.get_object("given_name"),
            self.top.get_object("nickname"),
            self.top.get_object("call"),
            self.top.get_object("suffix"),
            self.top.get_object("prefix"),
            self.top.get_object("surname"),
        ]:
            obj.connect("changed", self._changed_name)

        self.preview_name = self.top.get_object("full_name")
        self.surntab = SurnameTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_primary_name(),
            "person_editor_surnames",
            on_change=self._changed_name,
        )

    def get_start_date(self):
        """
        Get the start date for a person, usually a birth date, or
        something close to birth.
        """
        event = get_birth_or_fallback(self.dbstate.db, self.obj)
        return event.get_date_object() if event else None

    def get_end_date(self):
        """
        Get the end date for a person, usually a death date, or
        something close to death.
        """
        event = get_death_or_fallback(self.dbstate.db, self.obj)
        return event.get_date_object() if event else None

    def _create_tabbed_pages(self):
        """
        Create the notebook tabs and insert them into the main window.
        """
        notebook = Gtk.Notebook()
        notebook.set_scrollable(True)

        self.event_list = PersonEventEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj,
            "person_editor_events",
            start_date=self.get_start_date(),
            end_date=self.get_end_date(),
        )

        self._add_tab(notebook, self.event_list)
        self.track_ref_for_deletion("event_list")

        self.name_list = NameEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_alternate_names(),
            self.obj,
            "person_editor_names",
            self.name_callback,
        )
        self._add_tab(notebook, self.name_list)
        self.track_ref_for_deletion("name_list")

        self.srcref_list = CitationEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_citation_list(),
            "person_editor_citations",
            self.get_menu_title(),
        )
        self._add_tab(notebook, self.srcref_list)
        self.track_ref_for_deletion("srcref_list")

        self.attr_list = AttrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_attribute_list(),
            "person_editor_attributes",
        )
        self._add_tab(notebook, self.attr_list)
        self.track_ref_for_deletion("attr_list")

        self.addr_list = AddrEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_address_list(),
            "person_editor_addresses",
        )
        self._add_tab(notebook, self.addr_list)
        self.track_ref_for_deletion("addr_list")

        self.note_tab = NoteTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_note_list(),
            "person_editor_notes",
            self.get_menu_title(),
            notetype=NoteType.PERSON,
        )
        self._add_tab(notebook, self.note_tab)
        self.track_ref_for_deletion("note_tab")

        self.gallery_tab = GalleryTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_media_list(),
            self.load_person_image,
        )
        self._add_tab(notebook, self.gallery_tab)
        self.track_ref_for_deletion("gallery_tab")

        self.web_list = WebEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_url_list(),
            "person_editor_internet",
        )
        self._add_tab(notebook, self.web_list)
        self.track_ref_for_deletion("web_list")

        self.person_ref_list = PersonRefEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_person_ref_list(),
            "person_editor_associations",
        )
        self._add_tab(notebook, self.person_ref_list)
        self.track_ref_for_deletion("person_ref_list")

        self.lds_list = LdsEmbedList(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_lds_ord_list(),
            "person_editor_ldsord",
        )
        if not (config.get("interface.hide-lds") and self.lds_list.is_empty()):
            self._add_tab(notebook, self.lds_list)
        self.track_ref_for_deletion("lds_list")

        self.backref_tab = PersonBackRefList(
            self.dbstate,
            self.uistate,
            self.track,
            self.db.find_backlink_handles(self.obj.handle),
            "person_editor_references",
        )
        self._add_tab(notebook, self.backref_tab)
        self.track_ref_for_deletion("backref_tab")

        self._setup_notebook_tabs(notebook)
        notebook.show_all()
        self.top.get_object("vbox").pack_start(notebook, True, True, 0)

    def _changed_name(self, *obj):
        """
        callback to changes typed by user to the person name.
        Update the window title, and default name in name tab
        """
        self.update_title(self.get_menu_title())
        self.preview_name.set_markup(
            "<span size='x-large' weight='bold'>%s</span>"
            % markup_escape_text(self.get_preview_name(), -1)
        )
        self.name_list.update_defname()

    def name_callback(self):
        """
        Callback if changes happen in the name tab that impact the preferred
        name.
        """
        self.pname = self.obj.get_primary_name()

        self.ntype_field.reinit(self.pname.set_type, self.pname.get_type)
        self.given.reinit(self.pname.set_first_name, self.pname.get_first_name)
        self.call.reinit(self.pname.set_call_name, self.pname.get_call_name)
        self.title.reinit(self.pname.set_title, self.pname.get_title)
        self.suffix.reinit(self.pname.set_suffix, self.pname.get_suffix)
        self.nick.reinit(self.pname.set_nick_name, self.pname.get_nick_name)
        # part of Single Surname section
        self.surname_field.reinit(
            self.pname.get_primary_surname().set_surname,
            self.pname.get_primary_surname().get_surname,
        )

        self.prefix.reinit(
            self.pname.get_primary_surname().set_prefix,
            self.pname.get_primary_surname().get_prefix,
        )

        self.ortype_field.reinit(
            self.pname.get_primary_surname().set_origintype,
            self.pname.get_primary_surname().get_origintype,
        )

        self.__renewmultsurnames()

        if len(self.pname.get_surname_list()) == 1:
            self.singlesurn_active = True
        else:
            self.singlesurn_active = False
        if self.singlesurn_active:
            self.multsurnfr.hide()
            self.singsurnfr.show_all()
        else:
            self.singsurnfr.hide_all()
            self.multsurnfr.show_all()

    def build_menu_names(self, person):
        """
        Provide the information needed by the base class to define the
        window management menu entries.
        """
        return (_("Edit Person"), self.get_menu_title())

    def _image_button_press(self, obj, event):
        """
        Button press event that is caught when a button has been pressed while
        on the image on the main form.

        This does not apply to the images in galleries, just the image on the
        main form.

        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            media_list = self.obj.get_media_list()
            if media_list:
                media_ref = media_list[0]
                object_handle = media_ref.get_reference_handle()
                media_obj = self.db.get_media_from_handle(object_handle)

                try:
                    EditMediaRef(
                        self.dbstate,
                        self.uistate,
                        self.track,
                        media_obj,
                        media_ref,
                        self.load_photo,
                    )
                except WindowActiveError:
                    pass

        elif is_right_click(event):
            media_list = self.obj.get_media_list()
            if media_list:
                photo = media_list[0]
                self._show_popup(photo, event)
        # do not propagate further:
        return True

    def _show_popup(self, photo, event):
        """
        Look for right-clicks on a picture and create a popup menu of the
        available actions.
        """
        self.imgmenu = Gtk.Menu()
        menu = self.imgmenu
        obj = self.db.get_media_from_handle(photo.get_reference_handle())
        if obj:
            add_menuitem(menu, _("View"), photo, self._popup_view_photo)
        add_menuitem(
            menu, _("Edit Object Properties"), photo, self._popup_change_description
        )
        menu.popup_at_pointer(event)

    def _popup_view_photo(self, obj):
        """
        Open this picture in the default picture viewer.
        """
        media_list = self.obj.get_media_list()
        if media_list:
            photo = media_list[0]
            object_handle = photo.get_reference_handle()
            ref_obj = self.db.get_media_from_handle(object_handle)
            photo_path = media_path_full(self.db, ref_obj.get_path())
            open_file_with_default_application(photo_path, self.uistate)

    def _popup_change_description(self, obj):
        """
        Bring up the EditMediaRef dialog for the image on the main form.
        """
        media_list = self.obj.get_media_list()
        if media_list:
            media_ref = media_list[0]
            object_handle = media_ref.get_reference_handle()
            media_obj = self.db.get_media_from_handle(object_handle)
            EditMediaRef(
                self.dbstate,
                self.uistate,
                self.track,
                media_obj,
                media_ref,
                self.load_photo,
            )

    def _top_contextmenu(self, prefix):
        """
        Override from base class, the menuitems and actiongroups for the top
        of context menu.
        """
        if self.added:
            # Don't add items if not a real person yet
            return "", []

        _actions = [
            ("ActivePerson", self._make_active),
            ("HomePerson", self._make_home_person),
        ]

        ui_top_cm = (
            """
        <item>
          <attribute name="action">{prefix}.ActivePerson</attribute>
          <attribute name="label" translatable="yes">Make Active Person"""
            """</attribute>
        </item>
        <item>
          <attribute name="action">{prefix}.HomePerson</attribute>
          <attribute name="label" translatable="yes">Make Home Person"""
            """</attribute>
        </item>
        """.format(
                prefix=prefix
            )
        )

        return ui_top_cm, _actions

    def _top_drag_data_get(self, widget, context, sel_data, info, time):
        if info == DdTargets.PERSON_LINK.app_id:
            data = (DdTargets.PERSON_LINK.drag_type, id(self), self.obj.get_handle(), 0)
            sel_data.set(DdTargets.PERSON_LINK.atom_drag_type, 8, pickle.dumps(data))

    def _post_build_popup_ui(self):
        """
        Override base class, make inactive home action if not needed.
        """
        if self.added:
            return
        home_action = self.uistate.uimanager.get_action(self.action_group, "HomePerson")
        if (
            self.dbstate.db.get_default_person()
            and self.obj.get_handle()
            == self.dbstate.db.get_default_person().get_handle()
        ):
            home_action.set_enabled(False)
        else:
            home_action.set_enabled(True)

    def _make_active(self, obj, value):
        self.uistate.set_active(self.obj.get_handle(), "Person")

    def _make_home_person(self, obj, value):
        handle = self.obj.get_handle()
        if handle:
            self.dbstate.db.set_default_person_handle(handle)

    def _given_focus_out_event(self, entry, event):
        """
        Callback that occurs when the user leaves the given name field,
        allowing us to attempt to guess the gender of the person if
        so requested.
        """
        if not self.should_guess_gender:
            return False
        try:
            gender_type = self.db.genderStats.guess_gender(str(entry.get_text()))
            self.gender.force(gender_type)
        except:
            return False
        return False

    def _check_for_unknown_gender(self):
        if self.obj.get_gender() == Person.UNKNOWN:
            d = GenderDialog(parent=self.window)
            gender = d.run()
            d.destroy()
            if gender >= 0:
                self.obj.set_gender(gender)

    def _update_family_ids(self):
        # Update each of the families child lists to reflect any
        # change in ordering due to the new birth date
        family = self.obj.get_main_parents_family_handle()
        if family:
            f = self.db.get_family_from_handle(family)
            new_order = self.reorder_child_ref_list(self.obj, f.get_child_ref_list())
            f.set_child_ref_list(new_order)
        for family in self.obj.get_parent_family_handle_list():
            f = self.db.get_family_from_handle(family)
            new_order = self.reorder_child_ref_list(self.obj, f.get_child_ref_list())
            f.set_child_ref_list(new_order)

        error = False
        if self.original:
            (female, male, unknown) = _select_gender[self.obj.get_gender()]
            if male and self.original.get_gender() != Person.MALE:
                for tmp_handle in self.obj.get_family_handle_list():
                    temp_family = self.db.get_family_from_handle(tmp_handle)
                    if self.obj == temp_family.get_mother_handle():
                        if temp_family.get_father_handle() is not None:
                            error = True
                        else:
                            temp_family.set_mother_handle(None)
                            temp_family.set_father_handle(self.obj)
            elif female and self.original != Person.FEMALE:
                for tmp_handle in self.obj.get_family_handle_list():
                    temp_family = self.db.get_family_from_handle(tmp_handle)
                    if self.obj == temp_family.get_father_handle():
                        if temp_family.get_mother_handle() is not None:
                            error = True
                        else:
                            temp_family.set_father_handle(None)
                            temp_family.set_mother_handle(self.obj)
            elif unknown and self.original.get_gender() != Person.UNKNOWN:
                for tmp_handle in self.obj.get_family_handle_list():
                    temp_family = self.db.get_family_from_handle(tmp_handle)
                    if self.obj == temp_family.get_father_handle():
                        if temp_family.get_mother_handle() is not None:
                            error = True
                        else:
                            temp_family.set_father_handle(None)
                            temp_family.set_mother_handle(self.obj)
                    if self.obj == temp_family.get_mother_handle():
                        if temp_family.get_father_handle() is not None:
                            error = True
                        else:
                            temp_family.set_mother_handle(None)
                            temp_family.set_father_handle(self.obj)

            if error:
                msg2 = _("Problem changing the gender")
                msg = _(
                    "Changing the gender caused problems "
                    "with marriage information.\nPlease check "
                    "the person's marriages."
                )
                ErrorDialog(msg2, msg, parent=self.window)

    def save(self, *obj):
        """
        Save the data.
        """
        self.ok_button.set_sensitive(False)
        if self.object_is_empty():
            ErrorDialog(
                _("Cannot save person"),
                _(
                    "No data exists for this person. Please "
                    "enter data or cancel the edit."
                ),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return
        # fix surname problems
        for name in [self.obj.get_primary_name()] + self.obj.get_alternate_names():
            if len(name.get_surname_list()) > 1:
                newlist = [
                    surn for surn in name.get_surname_list() if not surn.is_empty()
                ]
                if len(newlist) != len(name.get_surname_list()):
                    name.set_surname_list(newlist)
            if len(name.get_surname_list()) == 0:
                name.set_surname_list([Surname()])
            try:
                primind = [
                    surn.get_primary() for surn in name.get_surname_list()
                ].index(True)
            except ValueError:
                primind = 0  # no match
            name.set_primary_surname(primind)

        # fix ide problems
        (uses_dupe_id, id) = self._uses_duplicate_id()
        if uses_dupe_id:
            prim_object = self.get_from_gramps_id(id)
            name = self.name_displayer.display(prim_object)
            msg1 = _("Cannot save person. ID already exists.")
            msg2 = _(
                "You have attempted to use the existing Gramps ID with "
                "value %(id)s. This value is already used by '"
                "%(prim_object)s'. Please enter a different ID or leave "
                "blank to get the next available ID value."
            ) % {"id": id, "prim_object": name}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        self._check_for_unknown_gender()

        self.db.set_birth_death_index(self.obj)

        if not self.obj.handle:
            with DbTxn(
                _("Add Person (%s)") % self.name_displayer.display(self.obj), self.db
            ) as trans:
                self.db.add_person(self.obj, trans)
        else:
            if self.data_has_changed():
                with DbTxn(
                    _("Edit Person (%s)") % self.name_displayer.display(self.obj),
                    self.db,
                ) as trans:
                    if not self.obj.get_gramps_id():
                        self.obj.set_gramps_id(self.db.find_next_person_gramps_id())
                    self.db.commit_person(self.obj, trans)

        self._do_close()
        if self.callback:
            self.callback(self.obj)
        self.callback = None

    def _edit_name_clicked(self, obj):
        """
        Bring up the EditName dialog for this name.

        Called when the edit name button is clicked for the primary name
        on the main form (not in the names tab).

        """
        try:
            EditName(
                self.dbstate, self.uistate, self.track, self.pname, self._update_name
            )
        except WindowActiveError:
            pass

    def _mult_surn_clicked(self, obj):
        """
        Show the list entry of multiple surnames
        """
        self.singsurnfr.hide_all()
        self.singlesurn_active = False
        # update multsurnfr for possible changes
        self.__renewmultsurnames()
        self.multsurnfr.show_all()

    def _update_name(self, name):
        """
        Called when the primary name has been changed by the EditName
        dialog.

        This allows us to update the main form in response to any changes.

        """
        for obj in (
            self.ntype_field,
            self.given,
            self.call,
            self.title,
            self.suffix,
            self.nick,
            self.surname_field,
            self.prefix,
            self.ortype_field,
        ):
            obj.update()

        self.__renewmultsurnames()

        if len(self.obj.get_primary_name().get_surname_list()) == 1:
            self.singlesurn_active = True
        else:
            self.singlesurn_active = False
        if self.singlesurn_active:
            self.multsurnfr.hide()
            self.singsurnfr.show_all()

        else:
            self.singsurnfr.hide_all()
            self.multsurnfr.show_all()

    def __renewmultsurnames(self):
        """Update mult surnames section with what is presently the
        correct surname.
        It is easier to recreate the entire mult surnames GUI than
        changing what has changed visually.
        """
        # remove present surname tab, and put new one
        msurhbox = self.top.get_object("hboxmultsurnames")
        msurhbox.remove(self.surntab)
        self.surntab = SurnameTab(
            self.dbstate,
            self.uistate,
            self.track,
            self.obj.get_primary_name(),
            "person_editor_surnames",
            on_change=self._changed_name,
        )
        self.multsurnfr.set_size_request(
            -1, int(config.get("interface.surname-box-height"))
        )
        msurhbox.pack_start(self.surntab, True, True, 0)

    def load_person_image(self):
        """
        Load the primary image into the main form if it exists.

        Used as callback on Gallery Tab too.

        """
        media_list = self.obj.get_media_list()
        if media_list:
            ref = media_list[0]
            handle = ref.get_reference_handle()
            obj = self.dbstate.db.get_media_from_handle(handle)
            if obj is None:
                # notify user of error
                from ..dialog import RunDatabaseRepair

                RunDatabaseRepair(
                    _("Non existing media found in the Gallery"), parent=self.window
                )
            else:
                self.load_photo(ref, obj)
        else:
            self.obj_photo.hide()
            self.frame_photo.hide()

    def load_photo(self, ref, obj):
        """
        Load the person's main photo using the Thumbnailer.
        """
        pixbuf = get_thumbnail_image(
            media_path_full(self.dbstate.db, obj.get_path()),
            obj.get_mime_type(),
            ref.get_rectangle(),
        )

        self.obj_photo.set_from_pixbuf(pixbuf)
        self.obj_photo.show()
        self.frame_photo.show_all()

    def birth_dates_in_order(self, child_ref_list):
        """
        Check any *valid* birthdates in the list to insure that they are in
        numerically increasing order.
        """
        inorder = True
        prev_date = 0
        handle_list = [ref.ref for ref in child_ref_list]
        for child_handle in handle_list:
            child = self.db.get_person_from_handle(child_handle)
            if child.get_birth_ref():
                event_handle = child.get_birth_ref().ref
                event = self.db.get_event_from_handle(event_handle)
                child_date = event.get_date_object().get_sort_value()
            else:
                continue
            if prev_date <= child_date:  # <= allows for twins
                prev_date = child_date
            else:
                inorder = False
        return inorder

    def reorder_child_ref_list(self, person, child_ref_list):
        """
        Reorder the child list to put the specified person in his/her
        correct birth order.

        Only check *valid* birthdates.  Move the person as short a distance as
        possible.

        """

        if self.birth_dates_in_order(child_ref_list):
            return child_ref_list

        # Build the person's date string once
        event_ref = person.get_birth_ref()
        if event_ref:
            event = self.db.get_event_from_handle(event_ref.ref)
            person_bday = event.get_date_object().get_sort_value()
        else:
            person_bday = 0

        # First, see if the person needs to be moved forward in the list
        handle_list = [ref.ref for ref in child_ref_list]

        index = handle_list.index(person.get_handle())
        target = index
        for i in range(index - 1, -1, -1):
            other = self.db.get_person_from_handle(handle_list[i])
            event_ref = other.get_birth_ref()
            if event_ref:
                event = self.db.get_event_from_handle(event_ref.ref)
                other_bday = event.get_date_object().get_sort_value()
                if other_bday == 0:
                    continue
                if person_bday < other_bday:
                    target = i
            else:
                continue

        # Now try moving to a later position in the list
        if target == index:
            for i in range(index, len(handle_list)):
                other = self.db.get_person_from_handle(handle_list[i])
                event_ref = other.get_birth_ref()
                if event_ref:
                    event = self.db.get_event_from_handle(event_ref.ref)
                    other_bday = event.get_date_object().get_sort_value()
                    if other_bday == "99999999":
                        continue
                    if person_bday > other_bday:
                        target = i
                else:
                    continue

        # Actually need to move?  Do it now.
        if target != index:
            ref = child_ref_list.pop(index)
            child_ref_list.insert(target, ref)
        return child_ref_list

    def _cleanup_on_exit(self):
        """Unset all things that can block garbage collection.
        Finalize rest
        """
        ##        self.private.destroy()
        ##        self.gender.destroy()
        ##        self.ntype_field.destroy()
        ##        self.given.destroy()
        ##        self.call.destroy()
        ##        self.title.destroy()
        ##        self.suffix.destroy()
        ##        self.nick.destroy()
        ##        self.surname_field.destroy()
        ##        self.prefix.destroy()
        ##        self.ortype_field.destroy()
        ##        self.tags.destroy()
        ##        self.gid.destroy()
        EditPrimary._cleanup_on_exit(self)
        # config.save()


class GenderDialog(Gtk.MessageDialog):
    def __init__(self, parent=None):
        Gtk.MessageDialog.__init__(
            self,
            transient_for=parent,
            modal=True,
            message_type=Gtk.MessageType.QUESTION,
        )
        self.set_icon(ICON)
        self.set_title("")

        self.set_markup(
            '<span size="larger" weight="bold">%s</span>'
            % _("Unknown gender specified")
        )
        self.format_secondary_text(
            _(
                "The gender of the person is currently unknown. "
                "Usually, this is a mistake. Please specify the gender."
            )
        )

        self.add_button(_("_Male"), Person.MALE)
        self.add_button(_("_Female"), Person.FEMALE)
        self.add_button(_("_Other"), Person.OTHER)
        self.add_button(_("_Unknown"), Person.UNKNOWN)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham, A. Roitman
# Copyright (C) 2007-2009  B. Malengier
# Copyright (C) 2008       Lukasz Rymarczyk
# Copyright (C) 2008       Raphael Ackermann
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2012       Doug Blank
# Copyright (C) 2012-2013  Paul Franklin
# Copyright (C) 2017       Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Module responsible for handling the command line arguments for Gramps.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import sys
import os
import getopt
import logging
import shutil
from glob import glob

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import (
    LONGOPTS,
    SHORTOPTS,
    USER_PLUGINS,
    VERSION_DIR,
    USER_CACHE,
    USER_CONFIG,
    USER_DATA,
    THUMB_DIR,
    USER_CSS,
)
from gramps.gen.utils.cast import get_type_converter
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

_HELP = _(
    """
Usage: gramps [OPTION...]
  --load-modules=MODULE1,MODULE2,...     Dynamic modules to load

Help options
  -?, --help                             Show this help message
  --usage                                Display brief usage message

Application options
  -O, --open=FAMILY_TREE                 Open Family Tree
  -U, --username=USERNAME                Database username
  -P, --password=PASSWORD                Database password
  -C, --create=FAMILY_TREE               Create on open if new Family Tree
  -i, --import=FILENAME                  Import file
  -e, --export=FILENAME                  Export file
  -r, --remove=FAMILY_TREE_PATTERN       Remove matching Family Tree(s) (use regular expressions)
  -f, --format=FORMAT                    Specify Family Tree format
  -a, --action=ACTION                    Specify action
  -p, --options=OPTIONS_STRING           Specify options
  -d, --debug=LOGGER_NAME                Enable debug logs
  -l [FAMILY_TREE_PATTERN...]            List Family Trees
  -L [FAMILY_TREE_PATTERN...]            List Family Trees in Detail
  -t [FAMILY_TREE_PATTERN...]            List Family Trees, tab delimited
  -u, --force-unlock                     Force unlock of Family Tree
  -s, --show                             Show config settings
  -c, --config=[config.setting[:value]]  Set config setting(s) and start Gramps
  -y, --yes                              Don't ask to confirm dangerous actions (non-GUI mode only)
  -q, --quiet                            Suppress progress indication output (non-GUI mode only)
  -v, --version                          Show versions
  -S, --safe                             Start Gramps in 'Safe mode'
                                          (temporarily use default settings)
  -D, --default=[APXFE]                  Reset settings to default;
                 A - addons are cleared
                 P - Preferences to default
                 X - Books are cleared, reports and tool settings to default
                 F - filters are cleared
                 E - Everything is set to default or cleared
"""
)

_USAGE = _(
    """
Example of usage of Gramps command line interface

1. To import four databases (whose formats can be determined from their names)
and then check the resulting database for errors, one may type:
gramps -i file1.ged -i file2.gpkg -i ~/db3.gramps -i file4.wft -a tool -p name=check.

2. To explicitly specify the formats in the above example, append filenames with appropriate -f options:
gramps -i file1.ged -f gedcom -i file2.gpkg -f gramps-pkg -i ~/db3.gramps -f gramps -i file4.wft -f wft -a tool -p name=check.

3. To record the database resulting from all imports, supply -e flag
(use -f if the filename does not allow Gramps to guess the format):
gramps -i file1.ged -i file2.gpkg -e ~/new-package -f gramps-pkg

4. To save any error messages of the above example into files outfile and errfile, run:
gramps -i file1.ged -i file2.dpkg -e ~/new-package -f gramps-pkg >outfile 2>errfile

5. To import three databases and start interactive Gramps session with the result:
gramps -i file1.ged -i file2.gpkg -i ~/db3.gramps

6. To open a database and, based on that data, generate timeline report in PDF format
putting the output into the my_timeline.pdf file:
gramps -O 'Family Tree 1' -a report -p name=timeline,off=pdf,of=my_timeline.pdf

7. To generate a summary of a database:
gramps -O 'Family Tree 1' -a report -p name=summary

8. Listing report options
Use the name=timeline,show=all to find out about all available options for the timeline report.
To find out details of a particular option, use show=option_name , e.g. name=timeline,show=off string.
To learn about available report names, use name=show string.

9. To convert a Family Tree on the fly to a .gramps xml file:
gramps -O 'Family Tree 1' -e output.gramps -f gramps

10. To generate a web site into an other locale (in german):
LANGUAGE=de_DE; LANG=de_DE.UTF-8 gramps -O 'Family Tree 1' -a report -p name=navwebpage,target=/../de

11. Finally, to start normal interactive session type:
gramps

Note: These examples are for bash shell.
Syntax may be different for other shells and for Windows.
"""
)


# -------------------------------------------------------------------------
# ArgParser
# -------------------------------------------------------------------------
class ArgParser:
    """
    This class is responsible for parsing the command line arguments (if any)
    given to gramps, and determining if a GUI or a CLI session must be started.
    A filename and/or options may be specified as arguments.

    The valid options are:

    -O, --open=FAMILY_TREE          Open Family Tree
    -U, --username=USERNAME         Database username
    -P, --password=PASSWORD         Database password
    -C, --create=FAMILY_TREE        Create on open if new Family Tree
    -i, --import=FILENAME           Import file
    -e, --export=FILENAME           Export file
    -r, --remove=PATTERN            Remove matching Family Tree(s)
    -f, --format=FORMAT             Specify Family Tree format
    -a, --action=ACTION             Specify action
    -p, --options=OPTIONS_STRING    Specify options
    -d, --debug=LOGGER_NAME         Enable debug logs
    -l [FAMILY_TREE...]             List Family Trees
    -L [FAMILY_TREE...]             List Family Trees in Detail
    -t [FAMILY_TREE...]             List Family Trees, tab delimited
    -u, --force-unlock              Force unlock of Family Tree
    -s, --show                      Show config settings
    -c, --config=SETTINGS           Set config setting(s) and start Gramps
    -y, --yes                       Don't ask to confirm dangerous actions
    -q, --quiet                     Suppress progress indication output
    -v, --version                   Show versions
    -h, --help                      Display the help
    --usage                         Display usage information

    If the filename (no options) is specified, the interactive session is
    launched using data from filename.  In this mode (filename, no options), the
    rest of the arguments are ignored.  This is a mode suitable by default for
    GUI launchers, mime type handlers, and the like.

    If no filename or -i option is given, a new interactive session (empty
    database) is launched, since no data is given anyway.

    If -O or -i option is given, but no -e or -a options are given, an
    interactive session is launched with the ``FILENAME`` (specified with -i).

    If both input (-O or -i) and processing (-e or -a) options are given,
    interactive session will not be launched.

    When using import or export options (-i or -e), the -f option may be
    specified to indicate the family tree format.

    Possible values for ``ACTION`` are:  'report', 'book' and 'tool'.

    Configuration ``SETTINGS`` may be specified using the -c option.  The
    settings are of the form config.setting[:value].  If used without a value,
    the setting is shown.

    If the -y option is given, the user's acceptance of any CLI prompt is
    assumed. (see :meth:`.cli.user.User.prompt`)

    If the -q option is given, extra noise on sys.stderr, such as progress
    indicators, is suppressed.
    """

    def __init__(self, args):
        """
        Pass the command line arguments on creation.
        """
        self.args = args

        self.open_gui = None
        self.open = None
        self.username = None
        self.password = None
        self.exports = []
        self.actions = []
        self.imports = []
        self.removes = []
        self.imp_db_path = None
        self.list = False
        self.list_more = False
        self.list_table = False
        self.database_names = None
        self.help = False
        self.usage = False
        self.force_unlock = False
        self.create = None
        self.quiet = False
        self.auto_accept = False

        self.errors = []
        self.parse_args()

    # -------------------------------------------------------------------------
    # Argument parser: sorts out given arguments
    # -------------------------------------------------------------------------
    def parse_args(self):
        """
        Fill in lists with open, exports, imports, and actions options.

        Any errors are added to self.errors
        """
        try:
            options, leftargs = getopt.getopt(self.args[1:], SHORTOPTS, LONGOPTS)
        except getopt.GetoptError as getopt_error:
            self.errors.append(
                self.construct_error(
                    "Type gramps --help for an overview of "
                    "commands, or read the manual pages.",
                    error=getopt_error,
                )
            )

            return

        # Some args can work on a list of databases:
        if leftargs:
            for option, value in options:
                if option in ["-L", "-l", "-t"]:
                    self.database_names = leftargs
                    leftargs = []

        if leftargs:
            # if there were an argument without option,
            # use it as a file to open and return
            self.open_gui = leftargs[0]
            print(_("Trying to open: %s ...") % leftargs[0], file=sys.stderr)
            # see if force open is on
            for option, value in options:
                if option in ("-u", "--force-unlock"):
                    self.force_unlock = True
                    break
            return

        # Go over all given option and place them into appropriate lists
        cleandbg = []
        need_to_quit = False
        for opt_ix, (option, value) in enumerate(options):
            if option in ["-O", "--open"]:
                self.open = value
            elif option in ["-C", "--create"]:
                self.create = value
            elif option in ["-U", "--username"]:
                self.username = value
            elif option in ["-P", "--password"]:
                self.password = value
            elif option in ["-i", "--import"]:
                family_tree_format = None
                if opt_ix < len(options) - 1 and options[opt_ix + 1][0] in (
                    "-f",
                    "--format",
                ):
                    family_tree_format = options[opt_ix + 1][1]
                self.imports.append((value, family_tree_format))
            elif option in ["-r", "--remove"]:
                self.removes.append(value)
            elif option in ["-e", "--export"]:
                family_tree_format = None
                if opt_ix < len(options) - 1 and options[opt_ix + 1][0] in (
                    "-f",
                    "--format",
                ):
                    family_tree_format = options[opt_ix + 1][1]
                abs_name = os.path.abspath(os.path.expanduser(value))
                if not os.path.exists(abs_name):
                    # The file doesn't exists, try to create it.
                    try:
                        open(abs_name, "w").close()
                        os.unlink(abs_name)
                    except OSError as e:
                        message = _(
                            "WARNING: %(strerr)s "
                            "(errno=%(errno)s):\n"
                            "WARNING: %(name)s\n"
                        ) % {"strerr": e.strerror, "errno": e.errno, "name": e.filename}
                        print(message)
                        sys.exit(1)
                self.exports.append((value, family_tree_format))
            elif option in ["-a", "--action"]:
                action = value
                if action not in ("report", "tool", "book"):
                    print(_("Unknown action: %s. Ignoring.") % action, file=sys.stderr)
                    continue
                options_str = ""
                if opt_ix < len(options) - 1 and options[opt_ix + 1][0] in (
                    "-p",
                    "--options",
                ):
                    options_str = options[opt_ix + 1][1]
                self.actions.append((action, options_str))
            elif option in ["-d", "--debug"]:
                print(_("setup debugging"), value, file=sys.stderr)
                logger = logging.getLogger(value)
                logger.setLevel(logging.DEBUG)
                cleandbg += [opt_ix]
            elif option in ["-l"]:
                self.list = True
            elif option in ["-L"]:
                self.list_more = True
            elif option in ["-t"]:
                self.list_table = True
            elif option in ["-s", "--show"]:
                from gramps.gen.config import config

                print(_("Gramps config settings from %s:") % config.filename)
                for sect, settings in config.data.items():
                    for settings_index, setting in settings.items():
                        print("%s.%s=%s" % (sect, settings_index, repr(value)))
                    print()
                sys.exit(0)
            elif option in ["-c", "--config"]:
                from gramps.gen.config import config

                cfg_name = value
                set_value = False
                if cfg_name:
                    if ":" in cfg_name:
                        cfg_name, new_value = cfg_name.split(":", 1)
                        set_value = True
                    if config.has_default(cfg_name):
                        setting_value = config.get(cfg_name)
                        print(
                            _("Current Gramps config setting: " "%(name)s:%(value)s")
                            % {"name": cfg_name, "value": repr(setting_value)},
                            file=sys.stderr,
                        )
                        if set_value:
                            # does a user want the default config value?
                            if new_value in ("DEFAULT", _("DEFAULT")):
                                new_value = config.get_default(cfg_name)
                            else:
                                converter = get_type_converter(setting_value)
                                new_value = converter(new_value)
                            config.set(cfg_name, new_value)
                            # Translators: indent "New" to match "Current"
                            print(
                                _(
                                    "    New Gramps config setting: "
                                    "%(name)s:%(value)s"
                                )
                                % {
                                    "name": cfg_name,
                                    "value": repr(config.get(cfg_name)),
                                },
                                file=sys.stderr,
                            )
                        else:
                            need_to_quit = True
                    else:
                        print(
                            _("Gramps: no such config setting: '%s'") % cfg_name,
                            file=sys.stderr,
                        )
                        need_to_quit = True
                cleandbg += [opt_ix]
            elif option in ["-h", "-?", "--help"]:
                self.help = True
            elif option in ["-u", "--force-unlock"]:
                self.force_unlock = True
            elif option in ["--usage"]:
                self.usage = True
            elif option in ["-y", "--yes"]:
                self.auto_accept = True
            elif option in ["-q", "--quiet"]:
                self.quiet = True
            elif option in ["-S", "--safe"]:
                cleandbg += [opt_ix]
            elif option in ["-D", "--default"]:

                def rmtree(path):
                    if os.path.isdir(path):
                        shutil.rmtree(path, ignore_errors=True)

                if "E" in value or "A" in value:  # clear addons
                    rmtree(USER_PLUGINS)
                if "E" in value or "P" in value:  # clear ini preferences
                    for fil in glob(os.path.join(VERSION_DIR, "*.*")):
                        if "custom_filters.xml" in fil:
                            continue
                        os.remove(fil)
                    # create gramps.ini so config won't load the one from an
                    # older version of Gramps.
                    with open(os.path.join(VERSION_DIR, "gramps.ini"), "w"):
                        pass
                if "E" in value or "F" in value:  # clear filters
                    fil = os.path.join(VERSION_DIR, "custom_filters.xml")
                    if os.path.isfile(fil):
                        os.remove(fil)
                if "E" in value or "X" in value:  # clear xml reports/tools
                    for fil in glob(os.path.join(USER_DATA, "*.xml")):
                        os.remove(fil)
                if "E" in value or "Z" in value:  # clear upgrade zips
                    for fil in glob(os.path.join(USER_DATA, "*.zip")):
                        os.remove(fil)
                if "E" in value:  # Everything else
                    rmtree(THUMB_DIR)
                    rmtree(USER_CSS)
                    rmtree(os.path.join(USER_CACHE, "maps"))
                    for fil in (
                        glob(os.path.join(USER_CACHE, "*"))
                        + glob(os.path.join(USER_CONFIG, "*"))
                        + glob(os.path.join(USER_DATA, "*"))
                    ):
                        if os.path.isfile(fil):
                            os.remove(fil)
                sys.exit(0)  # Done with Default

        # clean options list
        cleandbg.reverse()
        for ind in cleandbg:
            del options[ind]

        if (
            len(options) > 0
            and self.open is None
            and self.imports == []
            and self.removes == []
            and not (self.list or self.list_more or self.list_table or self.help)
        ):
            self.errors.append(
                self.construct_error(
                    "To use in the command-line mode, supply at "
                    "least one input file to process."
                )
            )

        if need_to_quit:
            sys.exit(0)

    def construct_error(self, suggestion_message, error=None):
        # Extract the arguments in the list.
        cli_args = "[ %s ]" % " ".join(self.args[1:])

        # The % operator replaces the list elements
        # with repr() of the list elements
        # which is OK for latin characters,
        # but not for non latin characters in list elements
        error_message = "Error parsing the arguments: %s \n"
        translated_message = _(error_message + suggestion_message) % cli_args

        if error:
            translated_message = str(error) + "\n" + translated_message

        return _("Error parsing the arguments"), translated_message

    # -------------------------------------------------------------------------
    # Determine the need for GUI
    # -------------------------------------------------------------------------
    def need_gui(self):
        """
        Determine whether we need a GUI session for the given tasks.
        """
        if self.errors:
            # errors in argument parsing ==> give cli error, no gui needed
            return False

        if len(self.removes) > 0:
            return False

        if self.list or self.list_more or self.list_table or self.help:
            return False

        if self.open_gui:
            # No-option argument, definitely GUI
            return True

        # If we have data to work with:
        if self.open or self.imports:
            if self.exports or self.actions:
                # have both data and what to do with it => no GUI
                return False
            elif self.create:
                if self.open:  # create an empty DB, open a GUI to fill it
                    return True
                else:  # create a DB, then do the import, with no GUI
                    self.open = self.create
                    return False
            else:
                # data given, but no action/export => GUI
                return True

        # No data, can only do GUI here
        return True

    def print_help(self):
        """
        If the user gives the --help or -h option, print the output to terminal.
        """
        if self.help:
            print(_HELP)
            sys.exit(0)

    def print_usage(self):
        """
        If the user gives the --usage print the output to terminal.
        """
        if self.usage:
            print(_USAGE)
            sys.exit(0)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008       Brian G. Matherly
# Copyright (C) 2008            Gary Burton
# Copyright (C) 2008            Robert Cheramy <robert@cheramy.net>
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Proxy class for the Gramps databases. Apply filter
"""

# -------------------------------------------------------------------------
#
# Gramps libraries
#
# -------------------------------------------------------------------------
from .proxybase import ProxyDbBase
from ..lib import (
    Date,
    Person,
    Name,
    Surname,
    NameOriginType,
    Family,
    Source,
    Citation,
    Event,
    Media,
    Place,
    Repository,
    Note,
    Tag,
)
from ..const import GRAMPS_LOCALE as glocale


class FilterProxyDb(ProxyDbBase):
    """
    A proxy to a Gramps database. This proxy will act like a Gramps database,
    but all data that does not match the provided filters will be hidden from
    the user.
    """

    def __init__(
        self, db, person_filter=None, event_filter=None, note_filter=None, user=None
    ):
        """
        Create a new FilterProxyDb instance.
        """
        ProxyDbBase.__init__(self, db)
        self.person_filter = person_filter
        if person_filter:
            self.plist = set(
                person_filter.apply(self.db, self.db.iter_person_handles(), user=user)
            )
        else:
            self.plist = set(self.db.iter_person_handles())

        if event_filter:
            self.elist = set(
                event_filter.apply(self.db, self.db.iter_event_handles(), user=user)
            )
        else:
            self.elist = set(self.db.iter_event_handles())

        if note_filter:
            self.nlist = set(
                note_filter.apply(self.db, self.db.iter_note_handles(), user=user)
            )
        else:
            self.nlist = set(self.db.iter_note_handles())

        self.flist = set()
        for handle in self.plist:
            person = self.db.get_person_from_handle(handle)
            if person:
                self.flist.update(person.get_family_handle_list())
                self.flist.update(person.get_parent_family_handle_list())

    def get_person_from_handle(self, handle):
        """
        Finds a Person in the database from the passed Gramps ID.
        If no such Person exists, None is returned.
        """
        if handle in self.plist:
            person = self.db.get_person_from_handle(handle)
            if person is None:
                return None
            person.set_person_ref_list(
                [ref for ref in person.get_person_ref_list() if ref.ref in self.plist]
            )

            person.set_family_handle_list(
                [hndl for hndl in person.get_family_handle_list() if hndl in self.flist]
            )

            person.set_parent_family_handle_list(
                [
                    hndl
                    for hndl in person.get_parent_family_handle_list()
                    if hndl in self.flist
                ]
            )

            eref_list = person.get_event_ref_list()
            bref = person.get_birth_ref()
            dref = person.get_death_ref()

            new_eref_list = [ref for ref in eref_list if ref.ref in self.elist]

            person.set_event_ref_list(new_eref_list)
            if bref in new_eref_list:
                person.set_birth_ref(bref)
            if dref in new_eref_list:
                person.set_death_ref(dref)

            # Filter notes out
            self.sanitize_person(person)

            return person
        else:
            return None

    def include_person(self, handle):
        return handle in self.plist

    def include_family(self, handle):
        return handle in self.flist

    def include_event(self, handle):
        return handle in self.elist

    def include_note(self, handle):
        return handle in self.nlist

    def get_source_from_handle(self, handle):
        """
        Finds a Source in the database from the passed Gramps ID.
        If no such Source exists, None is returned.
        """
        source = self.db.get_source_from_handle(handle)

        if source:
            # Filter notes out
            self.sanitize_notebase(source)

            media_ref_list = source.get_media_list()
            for media_ref in media_ref_list:
                self.sanitize_notebase(media_ref)
                attributes = media_ref.get_attribute_list()
                for attribute in attributes:
                    self.sanitize_notebase(attribute)

            repo_ref_list = source.get_reporef_list()
            for repo_ref in repo_ref_list:
                self.sanitize_notebase(repo_ref)

        return source

    def get_citation_from_handle(self, handle):
        """
        Finds a Citation in the database from the passed Gramps ID.
        If no such Citation exists, None is returned.
        """
        citation = self.db.get_citation_from_handle(handle)
        # Filter notes out
        self.sanitize_notebase(citation)
        return citation

    def get_media_from_handle(self, handle):
        """
        Finds a Media in the database from the passed Gramps handle.
        If no such Object exists, None is returned.
        """
        media = self.db.get_media_from_handle(handle)

        if media:
            # Filter notes out
            self.sanitize_notebase(media)

            attributes = media.get_attribute_list()
            for attr in attributes:
                self.sanitize_notebase(attr)

        return media

    def get_place_from_handle(self, handle):
        """
        Finds a Place in the database from the passed Gramps handle.
        If no such Place exists, None is returned.
        """
        place = self.db.get_place_from_handle(handle)

        if place:
            # Filter notes out
            self.sanitize_notebase(place)

            media_ref_list = place.get_media_list()
            for media_ref in media_ref_list:
                self.sanitize_notebase(media_ref)
                attributes = media_ref.get_attribute_list()
                for attribute in attributes:
                    self.sanitize_notebase(attribute)

        return place

    def get_event_from_handle(self, handle):
        """
        Finds a Event in the database from the passed Gramps ID.
        If no such Event exists, None is returned.
        """
        if handle in self.elist:
            event = self.db.get_event_from_handle(handle)
            # Filter all notes out
            self.sanitize_notebase(event)
            return event
        else:
            return None

    def get_family_from_handle(self, handle):
        """
        Finds a Family in the database from the passed Gramps ID.
        If no such Family exists, None is returned.
        """
        if handle in self.flist:
            family = self.db.get_family_from_handle(handle)
            if family is None:
                return None
            eref_list = [
                eref for eref in family.get_event_ref_list() if eref.ref in self.elist
            ]
            family.set_event_ref_list(eref_list)

            if family.get_father_handle() not in self.plist:
                family.set_father_handle(None)

            if family.get_mother_handle() not in self.plist:
                family.set_mother_handle(None)

            clist = [
                cref for cref in family.get_child_ref_list() if cref.ref in self.plist
            ]
            family.set_child_ref_list(clist)

            # Filter notes out
            for cref in clist:
                self.sanitize_notebase(cref)

            self.sanitize_notebase(family)

            attributes = family.get_attribute_list()
            for attr in attributes:
                self.sanitize_notebase(attr)

            event_ref_list = family.get_event_ref_list()
            for event_ref in event_ref_list:
                self.sanitize_notebase(event_ref)
                attributes = event_ref.get_attribute_list()
                for attribute in attributes:
                    self.sanitize_notebase(attribute)

            media_ref_list = family.get_media_list()
            for media_ref in media_ref_list:
                self.sanitize_notebase(media_ref)
                attributes = media_ref.get_attribute_list()
                for attribute in attributes:
                    self.sanitize_notebase(attribute)

            lds_ord_list = family.get_lds_ord_list()
            for lds_ord in lds_ord_list:
                self.sanitize_notebase(lds_ord)

            return family
        else:
            return None

    def get_repository_from_handle(self, handle):
        """
        Finds a Repository in the database from the passed Gramps ID.
        If no such Repository exists, None is returned.
        """
        repository = self.db.get_repository_from_handle(handle)
        # Filter notes out
        self.sanitize_notebase(repository)
        self.sanitize_addressbase(repository)
        return repository

    def get_note_from_handle(self, handle):
        """
        Finds a Note in the database from the passed Gramps ID.
        If no such Note exists, None is returned.
        """
        if handle in self.nlist:
            return self.db.get_note_from_handle(handle)
        else:
            return None

    def get_person_from_gramps_id(self, val):
        """
        Finds a Person in the database from the passed Gramps ID.
        If no such Person exists, None is returned.
        """
        person = self.db.get_person_from_gramps_id(val)
        if person:
            return self.get_person_from_handle(person.get_handle())
        else:
            return None

    def get_family_from_gramps_id(self, val):
        """
        Finds a Family in the database from the passed Gramps ID.
        If no such Family exists, None is returned.
        """
        family = self.db.get_family_from_gramps_id(val)
        if family:
            return self.get_family_from_handle(family.get_handle())
        else:
            return None

    def get_event_from_gramps_id(self, val):
        """
        Finds an Event in the database from the passed Gramps ID.
        If no such Event exists, None is returned.
        """
        event = self.db.get_event_from_gramps_id(val)
        if event:
            return self.get_event_from_handle(event.get_handle())
        else:
            return None

    def get_place_from_gramps_id(self, val):
        """
        Finds a Place in the database from the passed Gramps ID.
        If no such Place exists, None is returned.
        """
        place = self.db.get_place_from_gramps_id(val)
        if place:
            return self.get_place_from_handle(place.get_handle())
        else:
            return None

    def get_source_from_gramps_id(self, val):
        """
        Finds a Source in the database from the passed Gramps ID.
        If no such Source exists, None is returned.
        """
        source = self.db.get_source_from_gramps_id(val)
        if source:
            return self.get_source_from_handle(source.get_handle())
        else:
            return None

    def get_citation_from_gramps_id(self, val):
        """
        Finds a Citation in the database from the passed Gramps ID.
        If no such Citation exists, None is returned.
        """
        citation = self.db.get_citation_from_gramps_id(val)
        if citation:
            return self.get_citation_from_handle(citation.get_handle())
        else:
            return None

    def get_media_from_gramps_id(self, val):
        """
        Finds a Media in the database from the passed Gramps ID.
        If no such Media exists, None is returned.
        """
        media = self.db.get_media_from_gramps_id(val)
        if media:
            return self.get_media_from_handle(media.get_handle())
        else:
            return None

    def get_repository_from_gramps_id(self, val):
        """
        Finds a Repository in the database from the passed Gramps ID.
        If no such Repository exists, None is returned.
        """
        repository = self.db.get_repository_from_gramps_id(val)
        if repository:
            return self.get_repository_from_handle(repository.get_handle())
        else:
            return None

    def get_note_from_gramps_id(self, val):
        """
        Finds a Note in the database from the passed Gramps ID.
        If no such Note exists, None is returned.
        """
        note = self.db.get_note_from_gramps_id(val)
        if note:
            return self.get_note_from_handle(note.get_handle())
        else:
            return None

    def get_person_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Person in
        the database.

        :param sort_handles: If True, the list is sorted by surnames.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        # FIXME: plist is not a sorted list of handles
        return list(self.plist)

    def iter_person_handles(self):
        """
        Return an iterator over database handles, one handle for each Person in
        the database.
        """
        return self.plist

    def iter_people(self):
        """
        Return an iterator over objects for Persons in the database
        """
        return map(self.get_person_from_handle, self.plist)

    def get_event_handles(self):
        """
        Return a list of database handles, one handle for each Event in
        the database.
        """
        return list(self.elist)

    def iter_event_handles(self):
        """
        Return an iterator over database handles, one handle for each Event in
        the database.
        """
        return self.elist

    def iter_events(self):
        """
        Return an iterator over objects for Events in the database
        """
        return map(self.get_event_from_handle, self.elist)

    def get_family_handles(self, sort_handles=False, locale=glocale):
        """
        Return a list of database handles, one handle for each Family in
        the database.

        :param sort_handles: If True, the list is sorted by surnames.
        :type sort_handles: bool
        :param locale: The locale to use for collation.
        :type locale: A GrampsLocale object.
        """
        # FIXME: flist is not a sorted list of handles
        return list(self.flist)

    def iter_family_handles(self):
        """
        Return an iterator over database handles, one handle for each Family in
        the database.
        """
        return self.flist

    def iter_families(self):
        """
        Return an iterator over objects for Families in the database
        """
        return map(self.get_family_from_handle, self.flist)

    def get_note_handles(self):
        """
        Return a list of database handles, one handle for each Note in
        the database.
        """
        return list(self.nlist)

    def iter_note_handles(self):
        """
        Return an iterator over database handles, one handle for each Note in
        the database.
        """
        return self.nlist

    def iter_notes(self):
        """
        Return an iterator over objects for Notes in the database
        """
        return map(self.get_note_from_handle, self.nlist)

    def get_default_person(self):
        """returns the default Person of the database"""
        person = self.db.get_default_person()
        if person and person.get_handle() in self.plist:
            return person
        else:
            return None

    def get_default_handle(self):
        """returns the default Person of the database"""
        handle = self.db.get_default_handle()
        if handle in self.plist:
            return handle
        else:
            return None

    def has_person_handle(self, handle):
        """
        returns True if the handle exists in the current Person database.
        """
        return handle in self.plist

    def has_event_handle(self, handle):
        """
        returns True if the handle exists in the current Event database.
        """
        return handle in self.elist

    def has_family_handle(self, handle):
        """
        returns True if the handle exists in the current Family database.
        """
        return handle in self.flist

    def has_note_handle(self, handle):
        """
        returns True if the handle exists in the current Note database.
        """
        return handle in self.nlist

    def find_backlink_handles(self, handle, include_classes=None):
        """
        Find all objects that hold a reference to the object handle.
        Returns an iterator over a list of (class_name, handle) tuples.

        :param handle: handle of the object to search for.
        :type handle: database handle
        :param include_classes: list of class names to include in the results.
                                Default: None means include all classes.
        :type include_classes: list of class names

        This default implementation does a sequential scan through all
        the primary object databases and is very slow. Backends can
        override this method to provide much faster implementations that
        make use of additional capabilities of the backend.

        Note that this is a generator function, it returns a iterator for
        use in loops. If you want a list of the results use::

        >    result_list = list(find_backlink_handles(handle))
        """
        # FIXME: add a filter for returned handles (see private.py as an example)
        return self.db.find_backlink_handles(handle, include_classes)

    def sanitize_notebase(self, notebase):
        """
        Filters notes out of the passed notebase object according to the Note Filter.

        :param notebase: NoteBase object to clean
        :type event: NoteBase
        """
        if notebase:
            note_list = notebase.get_note_list()
            new_note_list = [note for note in note_list if note in self.nlist]
            notebase.set_note_list(new_note_list)

    def sanitize_addressbase(self, addressbase):
        if addressbase:
            addresses = addressbase.get_address_list()
            for address in addresses:
                self.sanitize_notebase(address)

    def sanitize_person(self, person):
        """
        Cleans filtered notes out of the passed person

        :param event: Person object to clean
        :type event: Person
        """
        if person:
            # Filter note references
            self.sanitize_notebase(person)
            self.sanitize_addressbase(person)

            name = person.get_primary_name()
            self.sanitize_notebase(name)

            altnames = person.get_alternate_names()
            for name in altnames:
                self.sanitize_notebase(name)

            self.sanitize_addressbase(person)

            attributes = person.get_attribute_list()
            for attr in attributes:
                self.sanitize_notebase(attr)

            event_ref_list = person.get_event_ref_list()
            for event_ref in event_ref_list:
                self.sanitize_notebase(event_ref)
                attributes = event_ref.get_attribute_list()
                for attribute in attributes:
                    self.sanitize_notebase(attribute)

            media_ref_list = person.get_media_list()
            for media_ref in media_ref_list:
                self.sanitize_notebase(media_ref)
                attributes = media_ref.get_attribute_list()
                for attribute in attributes:
                    self.sanitize_notebase(attribute)

            lds_ord_list = person.get_lds_ord_list()
            for lds_ord in lds_ord_list:
                self.sanitize_notebase(lds_ord)

            person_ref_list = person.get_person_ref_list()
            for person_ref in person_ref_list:
                self.sanitize_notebase(person_ref)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the basic functionality for a list view
"""

# -------------------------------------------------------------------------
#
# GTK
#
# -------------------------------------------------------------------------
from gi.repository import Pango, Gdk, Gtk, GdkPixbuf
from gramps.gen.const import THUMBSCALE

# -------------------------------------------------------------------------
#
# constants
#
# -------------------------------------------------------------------------
TEXT = 0
TOGGLE = 1
COMBO = 2
IMAGE = 3
INTEGER = 4
COLOR = 5

NOSORT = -1


# -------------------------------------------------------------------------
#
# ListModel
#
# -------------------------------------------------------------------------
class ListModel:
    """
    Simple model for lists in smaller dialogs (not views).

    tree:       A Gtk TreeView object.
    dlist:      A list of column definitions.  Each column definition is a tuple
                consisting of the following elements:
                (name, sort_id, width, type, editable, callback)
        name:       The column name.  If the name is an empty string then the
                    column is hidden.  Use a single space for the column to be
                    displayed but have no heading.
        sort_id:    The column id to used to sort the column.  Use the NOSORT
                    constant to disable sorting on the column.
        width:      The column width.
        type:       An optional column type.  One of the constants TEXT, TOGGLE,
                    COMBO, IMAGE, INTEGER or COLOR.  Default = TEXT.
        editable:   An optional boolean.  True if the column is editable.
                    Used with TEXT, INTEGER, COMBO and TOGGLE columns.
                    Default = False.
        callback:   An optional callback to be executed when the column is
                    edited. Used with TEXT, INTEGER, COMBO and TOGGLE columns.
                    Default = None.
    select_func:    Function called when the TreeView selection changes.
    event_func:     Function called when the user double-clicks on a row.
    mode:           Selection mode for TreeView.  See Gtk documentation.
    list_mode:      "list" or "tree"
    right_click:    Function called when the user right-clicks on a row.
    """

    def __init__(
        self,
        tree,
        dlist,
        select_func=None,
        event_func=None,
        mode=Gtk.SelectionMode.SINGLE,
        list_mode="list",
        right_click=None,
    ):
        self.tree = tree
        self.tree.set_fixed_height_mode(True)
        self.mylist = []
        self.data_index = 0
        self.sel_iter = None
        self.list_mode = list_mode  # "list", or "tree"
        self.double_click = None
        self.right_click = None

        for info in dlist:
            col_type = TEXT
            if isinstance(info, (list, tuple)):
                if len(info) > 3:
                    col_type = info[3]
            elif isinstance(info, dict):
                col_type = info.get("type", TEXT)
            # Now, add columns:
            if col_type == TOGGLE:
                self.mylist.append(bool)
            elif col_type == IMAGE:
                self.mylist.append(GdkPixbuf.Pixbuf)
            elif col_type == INTEGER:
                self.mylist.append(int)
            elif col_type == COLOR:
                self.mylist.append(str)
            else:
                self.mylist.append(str)
            self.data_index += 1
        self.mylist.append(object)

        self.function = {}
        self.model = None
        self.selection = None
        self.mode = mode
        self.new_model()
        self.count = 0
        self.cid = None
        self.cids = []
        self.idmap = {}

        self.__build_columns(dlist)
        self.connect_model()

        if select_func:
            self.selection.connect("changed", select_func)
        if event_func or right_click:
            if event_func:
                self.double_click = event_func
            if right_click:
                self.right_click = right_click
            self.tree.connect("button-press-event", self.__button_press)

    def __build_image_column(self, cnum, name, renderer, column):
        renderer = Gtk.CellRendererPixbuf()
        column = Gtk.TreeViewColumn(name[0], renderer)
        column.add_attribute(renderer, "pixbuf", cnum)
        renderer.set_property("height", THUMBSCALE / 2)
        return renderer, column

    def __build_columns(self, dlist):
        """
        Builds the columns based of the data in dlist
        """
        cnum = 0

        for item in dlist:
            visible_col = None
            if isinstance(item, (list, tuple)):
                if not item[2]:
                    continue
                else:
                    # [name, sort_id, width, [type, [editable, [callback]]]
                    name = item
            elif isinstance(item, dict):
                # valid fields: name, sort_id, width, type, editable, callback, visible_col
                name = [
                    item.get("name", " "),
                    item.get("sort_id", NOSORT),
                    item.get("width", 10),
                    item.get("type", TEXT),
                    item.get("editable", False),
                    item.get("callback", None),
                ]
                visible_col = item.get("visible_col", None)
            if len(name) == 3:
                name = (name[0], name[1], name[2], TEXT, False, None)
            elif len(name) == 4:
                name = (name[0], name[1], name[2], name[3], False, None)
            elif len(name) == 5:
                name = (name[0], name[1], name[2], name[3], name[4], None)

            if name[0] and name[3] == TOGGLE:
                renderer = Gtk.CellRendererToggle()
                if visible_col is not None:
                    column = Gtk.TreeViewColumn(name[0], renderer, visible=visible_col)
                else:
                    column = Gtk.TreeViewColumn(name[0], renderer)
                column.add_attribute(renderer, "active", cnum)
                if name[4]:
                    renderer.set_property("activatable", True)
                    renderer.connect("toggled", self.__toggled_cb, cnum)
                    if name[5]:
                        self.function[cnum] = name[5]
                else:
                    renderer.set_property("activatable", False)
            elif name[0] and name[3] == IMAGE:
                renderer, column = self.__build_image_column(
                    cnum, name, renderer, column
                )
            elif name[0] and name[3] == COLOR:
                renderer = Gtk.CellRendererText()
                if visible_col is not None:
                    column = Gtk.TreeViewColumn(
                        name[0], renderer, background=cnum, visible=visible_col
                    )
                else:
                    column = Gtk.TreeViewColumn(name[0], renderer, background=cnum)
            else:
                renderer = Gtk.CellRendererText()
                renderer.set_fixed_height_from_font(True)
                renderer.set_property("ellipsize", Pango.EllipsizeMode.END)
                if name[4]:
                    renderer.set_property("editable", True)
                    renderer.connect("edited", self.__edited_cb, cnum)
                    if name[5]:
                        self.function[cnum] = name[5]
                else:
                    renderer.set_property("editable", False)
                if visible_col is not None:
                    column = Gtk.TreeViewColumn(
                        name[0], renderer, text=cnum, visible=visible_col
                    )
                else:
                    column = Gtk.TreeViewColumn(name[0], renderer, text=cnum)
                column.set_reorderable(True)
            column.set_min_width(name[2])

            if name[0] == "":
                column.set_visible(False)
            else:
                column.set_resizable(True)
            if name[1] == -1:
                column.set_clickable(False)
            else:
                column.set_clickable(True)
                column.set_sort_column_id(name[1])

            column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
            # removed since we have a PersistentTreeView
            # column.set_fixed_width(name[2])

            cnum += 1
            self.cids.append(name[1])
            if name[0] != "":
                self.tree.append_column(column)

    def __toggled_cb(self, obj, path, col):
        """
        Callback executed when the checkbox of the cell renderer is clicked
        """
        new_value = not self.model[path][col]
        self.model[path][col] = new_value
        if col in self.function:
            self.function[col](path, new_value)

    def __edited_cb(self, cell, path, new_text, col):
        """
        Callback executed when the text of the cell renderer has changed
        """
        self.model[path][col] = new_text
        if col in self.function:
            self.function[col](path, new_text)

    def unselect(self):
        """
        Remove the selection from the view
        """
        self.selection.unselect_all()

    def set_reorderable(self, order):
        """
        Enables or disables reordering of data
        """
        self.tree.set_reorderable(order)

    def new_model(self):
        """
        Create a new model instance
        """
        if self.model:
            self.cid = self.model.get_sort_column_id()
            del self.model
            del self.selection
        self.count = 0
        if self.list_mode == "list":
            self.model = Gtk.ListStore(*self.mylist)
        elif self.list_mode == "tree":
            self.model = Gtk.TreeStore(*self.mylist)
        self.selection = self.tree.get_selection()
        self.selection.set_mode(self.mode)
        self.sel_iter = None

    def connect_model(self):
        """
        Connects the model to the associated tree
        """
        self.tree.set_model(self.model)
        if self.sel_iter:
            self.selection.select_iter(self.sel_iter)

        # if the sort column has not been defined (val[0] == -2), set
        # the sort column to the first column

        val = self.model.get_sort_column_id()
        if val[0] == -2 and self.cid:
            self.model.set_sort_column_id(self.cid[0], self.cid[1])
        self.sort()

    def sort(self):
        """
        Sorts the current view
        """
        val = self.model.get_sort_column_id()
        col = val[0]
        if col is None or col < 0:
            return
        if col > 0:
            self.model.set_sort_column_id(col, val[1])
        else:
            self.model.set_sort_column_id(self.cids[0], val[1])
        self.model.sort_column_changed()

    def get_selected(self):
        """
        Return the selected items
        """
        return self.selection.get_selected()

    def get_row_at(self, xpos, ypos):
        """
        Return the row at the specified (x,y) coordinates
        """
        path = self.tree.get_path_at_pos(xpos, ypos)
        if path is None:
            return self.count - 1
        else:
            return path[0][0] - 1

    def get_selected_row(self):
        """
        Get the selected row number
        """
        store, node = self.selection.get_selected()
        if node:
            rows = store.get_path(node).get_indices()
            return rows[0]
        else:
            return -1

    def get_selected_objects(self):
        """
        Return the list of selected objects in the list
        """
        if self.count == 0:
            return []
        elif self.mode == Gtk.SelectionMode.SINGLE:
            store, node = self.selection.get_selected()
            if node:
                return [self.model.get_value(node, self.data_index)]
            else:
                return []
        else:
            mlist = []
            self.selection.selected_foreach(self.__build_select_list, mlist)
            return mlist

    def get_icon(self):
        """
        Return an icon to be used for Drag and drop.
        """
        if self.mode == Gtk.SelectionMode.SINGLE:
            store, node = self.selection.get_selected()
            path = self.model.get_path(node)
        else:
            mlist = []
            self.selection.selected_foreach(self.__build_select_list, mlist)
            path = self.model.get_path(mlist[0])
        return self.tree.create_row_drag_icon(path)

    def __build_select_list(self, store, path, node, dlist):
        """
        GTK callback function for walking a select list
        """
        dlist.append(self.model.get_value(node, self.data_index))

    def clear(self):
        """
        Clears all data in the list
        """
        self.count = 0
        self.model.clear()

    def remove(self, node):
        """
        Remove the item from the model
        """
        self.model.remove(node)
        self.count -= 1

    def get_row(self, node):
        """
        Return the row associated with the selected node
        """
        row = self.model.get_path(node)
        return row[0]

    def select_row(self, row):
        """
        Selects the item based on path
        """
        self.selection.select_path(row)

    def select_iter(self, node):
        """
        Selects the item based on iter
        """
        self.selection.select_iter(node)

    def get_object(self, node):
        """
        Return the object associated with the node. This is controlled
        by extracting the data from the associated data index
        """
        return self.model.get_value(node, self.data_index)

    def insert(self, position, data, info=None, select=0):
        """
        Inserts the item at the specified position in the model.
        """
        self.count += 1
        node = self.model.insert(position)
        col = 0
        for obj in data:
            self.model.set_value(node, col, obj)
            col += 1
        if info:
            self.model.set_value(node, col, info)
            self.idmap[str(info)] = node
        if select:
            self.selection.select_iter(node)
        return node

    def get_data(self, node, cols):
        """
        Return a list of data from the model associated with the node
        """
        return [self.model.get_value(node, c) for c in cols]

    def add(self, data, info=None, select=0, node=None):
        """
        Add the data to the model at the end of the model
        """
        info = info or ""
        self.count += 1
        need_to_set = True
        # Create the node:
        if self.list_mode == "list":
            node = self.model.append()
        elif self.list_mode == "tree":
            if node is None:  # new node
                node = self.model.append(None, data + [info])
                need_to_set = False
            else:  # use a previous node, passed in
                node = self.model.append(node)
        # Add data:
        if need_to_set:
            col = 0
            for obj in data:
                self.model.set_value(node, col, obj)
                col += 1
            self.model.set_value(node, col, info)
        # Set info and select:
        if info is not None:
            self.idmap[str(info)] = node
        if select:
            self.sel_iter = node
            self.selection.select_iter(self.sel_iter)
        # Return created node
        return node

    def set(self, node, data, info=None, select=0):
        """
        Change the data associated with the specific node. It does not
        add any data, just alters an existing row.
        """
        col = 0
        for obj in data:
            self.model.set_value(node, col, obj)
            col += 1
        self.model.set_value(node, col, info)
        if info:
            self.idmap[str(info)] = node
        if select:
            self.sel_iter = node
        return node

    def __button_press(self, obj, event):
        """
        Called when a button press is executed
        """
        from .utils import is_right_click

        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            if self.double_click:
                self.double_click(obj)
                return True
        elif is_right_click(event):
            if self.right_click:
                self.right_click(obj, event)
                return True
        return False

    def find(self, info):
        """
        Selects the item associated with the pass information.
        """
        if info in self.idmap:
            node = self.idmap[str(info)]
            self.selection.select_iter(node)

    def move_up(self, row):
        """
        Move the given row up one position.
        """
        if row < 1 or row == -1:
            return False
        this_row = self.model.get_iter((row,))
        prev_row = self.model.get_iter((row - 1,))
        self.model.move_before(this_row, prev_row)
        return True

    def move_down(self, row):
        """
        Move the given row down one position.
        """
        if row >= self.count - 1 or row == -1:
            return False
        this_row = self.model.get_iter((row,))
        next_row = self.model.get_iter((row + 1,))
        self.model.move_after(this_row, next_row)
        return True
